"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp import QuantumVariable, x
from qrisp.misc import gate_wrap


@gate_wrap(is_qfree=True)
def increment(qf, amount):
    if amount == 0:
        return

    if not abs(amount) in [qf.decoder(i) for i in range(2**qf.size)]:
        raise Exception("Tried to increment with invalid value")

    # Convert amount to integer reesentation
    amount = amount / 2**qf.exponent

    if amount < 0:
        neg_value = True
        amount = -amount
    else:
        neg_value = False

    # Factorize amount such that amount = amount_tilde 2**power_offset
    # This implies that amount has a binary representation like
    # 10101100000000000
    # where the amount of zeros is the power offset
    # Addition on these significances dont need any further treatment
    power_offset = 0

    while amount % 2 == 0:
        power_offset += 1
        amount = amount / 2

    # Prepare qubit list
    qubit_list = qf.reg[power_offset:]

    if neg_value:
        x(qf.reg)

    amount = int(amount)
    if amount == 1:
        increment_qb_list(qf.qs, qubit_list)
    else:
        increment_arbitrary_constant_qubit_list(qf.qs, qubit_list, amount)

    if neg_value:
        x(qf.reg)


# Incrementation method presented at
# https://algassert.com/circuits/2015/06/12/Constructing-Large-Increment-Gates.html
def increment_qb_list(qs, qubit_list):
    n = len(qubit_list)

    if n == 1:
        qs.x(qubit_list[0])
        return
    if n == 2:
        qs.cx(qubit_list[0], qubit_list[1])
        qs.x(qubit_list[0])
        return

    incr_anc = QuantumVariable(n - 2)
    qs.mcx([qubit_list[0], qubit_list[1]], incr_anc[0], method="gray_pt")

    for i in range(2, n - 1):
        qs.mcx([qubit_list[i], incr_anc[i - 2]], incr_anc[i - 1], method="gray_pt")

    qs.cx(incr_anc[-1], qubit_list[-1])

    for i in range(n - 2, 1, -1):
        qs.mcx([qubit_list[i], incr_anc[i - 2]], incr_anc[i - 1], method="gray_pt_inv")
        qs.cx(incr_anc[i - 2], qubit_list[i])

    qs.mcx([qubit_list[0], qubit_list[1]], incr_anc[0], method="gray_pt_inv")
    qs.cx(qubit_list[0], qubit_list[1])
    qs.x(qubit_list[0])

    incr_anc.delete()


# Incrementation circuit based on the Cuccaro adder
# The idea is to perform the Cuccaro adder but instead of performing
# CNOTs conditioned on the classically known incrementation qubits, we perform NOT gates
# classically conditioned on the incrementation


# For this we define the semi-classical versions of the MAJ and UMA gates depending
# on the (classically known) value of the a variables from the Cuccaro construction
def cMAJ_gate(val):
    from qrisp import QuantumCircuit

    qc = QuantumCircuit(3)

    if val:
        qc.x(0)
        qc.x(1)

    # qc.mcx([0,1], 2, method = "gray_pt")
    qc.mcx([0, 1], 2)

    result = qc.to_gate()

    result.name = "cMAJ"

    return result


def cUMA_gate(val):
    from qrisp import QuantumCircuit

    qc = QuantumCircuit(3)

    # qc.mcx([0,1], 2, method = "gray_pt_inv")
    qc.mcx([0, 1], 2)
    if val:
        qc.x(0)
    qc.cx(0, 1)

    result = qc.to_gate()
    result.name = "cUMA" + str(val)
    return result


# Incrementation by an arbitrary constant using the Cuccaro Adder
# ie. this performs
# qf += x
def increment_arbitrary_constant_qubit_list(qs, qubit_list, c):
    n = len(qubit_list)

    # Initialize the a variable
    from qrisp.core import QuantumVariable

    incr_ancilla = QuantumVariable(n)

    from qrisp.misc import int_as_array

    # Calculate the (classically known) bitstring of c

    a = int_as_array(c, n)[::-1]

    for i in range(n):
        if a[i]:
            x(incr_ancilla[i])

    incr_carry_ancilla = QuantumVariable(1)

    # Perform modular Cuccaro Addition

    qubit_list_1 = incr_ancilla.reg[:-1]
    qubit_list_2 = qubit_list[:-1]

    # Prepare MAJ/UMA gate qubits
    slot_1_qbs = [incr_carry_ancilla.reg[0]] + qubit_list_1[:-1]
    slot_2_qbs = qubit_list_2
    slot_3_qbs = qubit_list_1

    output_qubit = qubit_list[-1]

    iterations = len(slot_1_qbs)

    # Perform 1st step of the modular addition section
    for i in range(iterations):
        qbits = [slot_1_qbs[i], slot_2_qbs[i], slot_3_qbs[i]]
        qs.append(cMAJ_gate(a[i]), qbits)

    # Calculate output qbit
    qs.cx(qbits[2], output_qubit)

    # Perform UMA iterations
    for i in range(iterations - 1, -1, -1):
        qbits = [slot_1_qbs[i], slot_2_qbs[i], slot_3_qbs[i]]
        qs.append(cUMA_gate(a[i]), qbits)

    # Perform the 2nd step for achieving modular arithmetic
    qs.cx(incr_ancilla.reg[-1], output_qubit)

    # Uncompute the ancilla variable
    for i in range(n):
        if a[i]:
            x(incr_ancilla[i])

    # Delete ancilla

    incr_ancilla.delete()
    incr_carry_ancilla.delete()



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 23.05.2022
import numpy as np

from qrisp.core import QuantumSession
from qrisp.interface.backends import VirtualBackend
from qrisp.interface.backends import VirtualQiskitBackend
from qiskit import Aer


def test_qiskit_backend_client():
    # TO-DO prevent this test from crashing regardless of functionality
    return
    # Create QuantumSession
    qs = QuantumSession()

    qs.add_qubit()
    qs.add_qubit()
    qs.add_clbit()

    qs.h(0)

    qs.rz(np.pi / 2, 0)

    qs.x(0)
    qs.cx(0, 1)
    qs.measure(1, 0)

    qs.append(qs.to_op("composed_op"), qs.qubits, qs.clbits)

    qs.append(qs.to_op("multi_composed_op"), qs.qubits, qs.clbits)

    print(qs)

    ###################

    # Create VirtualBackend
    def sample_run_func(qc, shots):
        print("Executing Circuit")
        return {"0": shots}

    test_virtual_backend = VirtualBackend(sample_run_func)

    print(test_virtual_backend.run(qs, 100))
    assert str(test_virtual_backend.run(qs, 100)) == "{'0': 100}"
    assert test_virtual_backend.run(qs, 100)["0"] == 100

    ###################

    # Create Qiskit Backend
    test_qiskit_backend = VirtualQiskitBackend(Aer.get_backend("qasm_simulator"))

    from qrisp import QuantumCircuit

    qc = QuantumCircuit(4, 1)
    qc.x(0)
    qc.measure(0, 0)

    print(test_qiskit_backend.run(qc, 2000))
    # status = test_qiskit_backend.ping()
    assert str(test_qiskit_backend.run(qc, 2000)) == "{'1': 2000}"



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.core.library import mcx, p, rz, x, z
from qrisp.core.quantum_variable import QuantumVariable
from qrisp.core.session_merging_tools import merge
from qrisp.environments.quantum_environments import QuantumEnvironment
from qrisp.environments.quantum_inversion import invert
from qrisp.misc import (
    find_calling_line,
    perm_lock,
    perm_unlock,
    redirect_qfunction,
    unlock,
)


def quantum_condition(function):
    def q_env_generator(*args, **kwargs):
        return ConditionEnvironment(function, args, kwargs=kwargs)

    return q_env_generator


# Class to describe if-environments
# For more information on how Environments work, check the QuantumEnvironment class.
# This class is instantiated with a function returning a qubit,
# which indicates wther the condition is met.
# On entering this environment, the function will be evaluated and the content
# (i.e. the quantum operations) will be collected.
# When compiled, this environment will check for parent conditional environments and
# include their truth value into the compilation process. Subsequently, every operation
# that happened inside this environment is turned into it's controlled version.
# Finally, the function evaluating the truth value will be uncomputed and
# the environment is reset for it's next use.
class ConditionEnvironment(QuantumEnvironment):
    r"""

    This class enables the usage of *if*-conditionals as we are used to from
    classical programming: ::

        from qrisp import QuantumChar, QuantumFloat, h, multi_measurement

        q_ch = QuantumChar()
        qf = QuantumFloat(3, signed = True)

        h(q_ch[0])

        with q_ch == "a":
            qf += 2


    >>> print(multi_measurement([q_ch,qf]))
    {('a', 2): 0.5, ('b', 0): 0.5}

    In this code snippet, we first bring the :ref:`QuantumChar` ``q_ch`` into the
    superposition

    .. math::

        \ket{\text{q_ch}} = \frac{1}{\sqrt{2}} \left( \ket{\text{"a"}}
        + \ket{\text{"b"}} \right)

    After that, we enter the ConditionEnvironment, which controls the operations
    on the condition that ``q_ch`` is in state $\ket{a}$. Finally, we simultaneously
    measure both :ref:`QuantumVariables <QuantumVariable>`. We see that the
    incrementation of ``qf`` only occured on the branch, where ``q_ch`` is equal to the
    character ``"a"``. The resulting quantum state is:

    .. math::

        \ket{\psi} = \frac{1}{\sqrt{2}} \left( \ket{\text{"a"}}\ket{2}
        + \ket{\text{"b"}}\ket{0} \right)

    It is furthermore possible to invert the condition truth value or apply phases.
    For this we acquire the :ref:`QuantumBool` containing the truth value using the *as*
    statement ::

        from qrisp import x, p
        import numpy as np

        with q_ch == "a" as cond_bool:
            qf += 2
            cond_bool.flip()
            qf -= 2
            p(np.pi/4, cond_bool)

    >>> qf.qs.statevector()
    sqrt(2)*(|a>*|4> + exp(I*pi/4)*|b>*|-2>)/2

    **Constructing custom conditional environments**

    Apart from infix notation like the equality operator, Qrisp also provides an
    interface for creating custom conditonal environments.

    Parameters
    ----------
    cond_eval_function : function
        A function which evaluates the truth value. Must return a :ref:`QuantumBool`.
        Intermediate results do not have to be uncomputed or deleted, as this is
        automatically performed, when the condition truth value is uncomputed.
    args : list
        The arguments on which to evaluate.
    kwargs : dict, optional
        A dictionary of keyword arguments for ``cond_eval_function``. The default is {}.


    Examples
    --------

    We will now demonstrate how a ConditionEnvironment, that evaluates the equality of
    two :ref:`QuantumVariables <QuantumVariable>` can be constructed: ::

        from qrisp import QuantumBool, QuantumVariable, x, cx, mcx

        def quantum_eq(qv_0, qv_1):



            if qv_0.size != qv_1.size:
                raise Exception("Tried to evaluate equality condition for
                QuantumVariables of differing size")

            temp_qv = QuantumVariable(qv_0.size)

            cx(qv_0, temp_qv)
            cx(qv_1, temp_qv)
            x(temp_qv)

            res = QuantumBool()

            mcx(temp_qv, res)

            return res

    In this function, we create a temporary variable where we apply CX gates controlled
    on the inputs onto. The qubits where ``qv_0`` and ``qv_1`` agree, will then be in
    state $\ket{0}$. After this, we apply regular X gates onto ``temp_qv`` such that
    the qubits where the inputs agree, are in state $\ket{1}$. Finally,
    we apply a multi-controlled X gate onto the result to flip the result, if
    all of qubits of the qubits in ``temp_qv`` are in the $\ket{0}$ state.

    We inspect the resulting QuantumCircuit

    >>> from qrisp import QuantumChar
    >>> q_ch_0 = QuantumChar()
    >>> q_ch_1 = QuantumChar()
    >>> res_bool = quantum_eq(q_ch_0, q_ch_1)
    >>> print(q_ch_0.qs)
    QuantumCircuit:
    ---------------
     q_ch_0.0: ──■─────────────────────────────────────────────────────────
                 │
     q_ch_0.1: ──┼────■────────────────────────────────────────────────────
                 │    │
     q_ch_0.2: ──┼────┼────■───────────────────────────────────────────────
                 │    │    │
     q_ch_0.3: ──┼────┼────┼────■──────────────────────────────────────────
                 │    │    │    │
     q_ch_0.4: ──┼────┼────┼────┼────■─────────────────────────────────────
                 │    │    │    │    │
     q_ch_1.0: ──┼────┼────┼────┼────┼────■────────────────────────────────
                 │    │    │    │    │    │
     q_ch_1.1: ──┼────┼────┼────┼────┼────┼────■───────────────────────────
                 │    │    │    │    │    │    │
     q_ch_1.2: ──┼────┼────┼────┼────┼────┼────┼────■──────────────────────
                 │    │    │    │    │    │    │    │
     q_ch_1.3: ──┼────┼────┼────┼────┼────┼────┼────┼────■─────────────────
                 │    │    │    │    │    │    │    │    │
     q_ch_1.4: ──┼────┼────┼────┼────┼────┼────┼────┼────┼────■────────────
               ┌─┴─┐  │    │    │    │  ┌─┴─┐  │    │    │    │  ┌───┐
    temp_qv.0: ┤ X ├──┼────┼────┼────┼──┤ X ├──┼────┼────┼────┼──┤ X ├──■──
               └───┘┌─┴─┐  │    │    │  └───┘┌─┴─┐  │    │    │  ├───┤  │
    temp_qv.1: ─────┤ X ├──┼────┼────┼───────┤ X ├──┼────┼────┼──┤ X ├──■──
                    └───┘┌─┴─┐  │    │       └───┘┌─┴─┐  │    │  ├───┤  │
    temp_qv.2: ──────────┤ X ├──┼────┼────────────┤ X ├──┼────┼──┤ X ├──■──
                         └───┘┌─┴─┐  │            └───┘┌─┴─┐  │  ├───┤  │
    temp_qv.3: ───────────────┤ X ├──┼─────────────────┤ X ├──┼──┤ X ├──■──
                              └───┘┌─┴─┐               └───┘┌─┴─┐├───┤  │
    temp_qv.4: ────────────────────┤ X ├────────────────────┤ X ├┤ X ├──■──
                                   └───┘                    └───┘└───┘┌─┴─┐
        res.0: ───────────────────────────────────────────────────────┤ X ├
                                                                      └───┘
    Live QuantumVariables:
    ----------------------
    QuantumChar q_ch_0
    QuantumChar q_ch_1
    QuantumVariable temp_qv
    QuantumBool res

    We can now construct the conditional environment from this function ::

        from qrisp import ConditionEnvironment, multi_measurement, h

        #Create some sample arguments on which to evaluate the condition

        q_bool_0 = QuantumBool()
        q_bool_1 = QuantumBool()
        q_bool_2 = QuantumBool()

        h(q_bool_0)

        with ConditionEnvironment(cond_eval_function = quantum_eq,
                                  args = [q_bool_0, q_bool_1]):
            q_bool_2.flip()

    >>> print(multi_measurement([q_bool_0, q_bool_1, q_bool_2]))
    {(False, False, True): 0.5, (True, False, False): 0.5}

    This agrees with our expectation, that ``q_bool_2`` is only ``True`` if the other
    two agree.

    **The quantum_condition decorator**

    Creating quantum conditions like this seems a bit unwieldy.
    For a more convenient solution, we provide the ``quantum_condition`` decorator.
    This decorator can be applied to a function returning a :ref:`QuantumBool`,
    which is then returning the corresponding ConditionEnvironment instead.
    To demonstrate, we construct a "less than" condition for QuantumFloats ::

        from qrisp import quantum_condition, cx

        @quantum_condition
        def less_than(qf_0, qf_1):

            temp_qf = qf_0 - qf_1

            res = QuantumBool()

            cx(temp_qf.sign(), res)

            return res

        qf_0 = QuantumFloat(3)
        qf_1 = qf_0.duplicate()

        qf_0[:] = 2
        h(qf_1[:2])

        res_q_bool = QuantumBool()

        with less_than(qf_0, qf_1):
            res_q_bool.flip()

    >>> print(multi_measurement([qf_0, qf_1, res_q_bool]))
    {(2, 0, False): 0.25, (2, 1, False): 0.25, (2, 2, False): 0.25, (2, 3, True): 0.25}


    **Quantum-Loops**

    An interesting application of conditional environments is the ``qRange`` iterator.
    Using this construct, we can mimic a loop as we are used from classical computing,
    where the end of the loop is determined by a quantum state: ::

        from qrisp import QuantumFloat, qRange, h

        n = QuantumFloat(3)
        qf = QuantumFloat(5)

        n[:] = 4

        h(n[0])

        n_results = n.get_measurement()

        for i in qRange(n):
            qf += i


    >>> print(qf)
    {10: 0.5, 15: 0.5}

    This script calculates the sum of all integers up to a certain threshold.
    The threshold (n) is a :ref:`QuantumFloat` in superposition, implying the result of
    the sum is also in a superposition. The expected results can be quickly determined
    by using Gauß's formula:

    .. math::

        \sum_{i = 0}^n i = \frac{n(n+1)}{2}


    >>> print("Excpected outcomes:", [n*(n+1)/2 for n in n_results.keys()])
    Excpected outcomes: [10.0, 15.0]

    """

    # Constructor of the class

    def __init__(self, cond_eval_function, args, kwargs={}):
        # The function which evaluates the condition - should return a QuantumBool
        self.cond_eval_function = cond_eval_function

        # Save the arguments on which the function should be evaluated
        self.args = args

        # Save the keyword arguments
        self.kwargs = kwargs

        # Note the QuantumSession of the arguments of the arguments

        self.arg_qs = merge(args)

        self.manual_allocation_management = True

    # Method to enter the environment
    def __enter__(self):
        from qrisp.qtypes.quantum_bool import QuantumBool

        # For more information on why this attribute is neccessary check the comment
        # on the line containing subcondition_truth_values = []
        self.sub_condition_envs = []

        self.qbool = QuantumBool(name="cond_env*", qs=self.arg_qs)
        self.condition_truth_value = self.qbool[0]

        super().__enter__()

        merge(self.env_qs, self.arg_qs)
        return self.qbool

    def __exit__(self, exception_type, exception_value, traceback):
        from qrisp.environments import ControlEnvironment, InversionEnvironment

        # We determine the parent condition environment
        self.parent_cond_env = None

        QuantumEnvironment.__exit__(self, exception_type, exception_value, traceback)

        for env in self.env_qs.env_stack[::-1]:
            if isinstance(env, (ConditionEnvironment, ControlEnvironment)):
                self.parent_cond_env = env
                break
            if not isinstance(env, (QuantumEnvironment, InversionEnvironment)):
                break

    # Compile method
    def compile(self):
        from qrisp.environments import ControlEnvironment
        from qrisp.qtypes.quantum_bool import QuantumBool

        # Create the quantum variable where the condition truth value should be saved
        # Incase we have a parent environment we create two qubits because
        # we use the second qubit to compute the toffoli of this one and the parent
        # environments truth value in order to not have the environment operations
        # controlled on two qubits
        if len(self.env_data) == 0:
            self.qbool.delete()
        else:
            # The first step we have to perform is calculating the truth value of
            # this environments quantum condition. For this we differentiate between
            # the case that this condition is embedded in another condition or not

            if self.parent_cond_env is not None:
                # In the parent case we also need to make sure that the code is executed
                # if the parent environment is executed. For this a possible approach
                # would be to control the content on both, the parent and
                # the chield truth value. However, for each nesting level the gate count
                # to generate the controlled-controlled-controlled... version
                # of the gates inside the environment increases exponentially.
                # Because of this we compute the toffoli of the parent and
                # child truth value and control the environment gates on this qubit.

                cond_eval_bool = self.cond_eval_function(*self.args, **self.kwargs)

                if not isinstance(cond_eval_bool, QuantumBool):
                    raise Exception(
                        "Tried to compile QuantumCondition environment with"
                        "a condition evaluation function not returning a QuantumBool"
                    )

                # Create and execute phase tolerant toffoli gate
                toffoli_qb_list = [
                    self.parent_cond_env.condition_truth_value,
                    cond_eval_bool,
                ]

                from qrisp import mcx

                mcx(toffoli_qb_list, self.condition_truth_value, method="gray_pt")

            else:
                # Without any parent environment we can simply synhesize the
                # truth value of the quantum condition into it's qubit

                redirect_qfunction(self.cond_eval_function)(
                    *self.args, target=self.qbool, **self.kwargs
                )

                if isinstance(self.env_qs.data[-1], QuantumEnvironment):
                    env = self.env_qs.data.pop(-1)
                    env.compile()

                cond_eval_bool = self.qbool

            from qrisp import recursive_qv_search

            perm_lock(recursive_qv_search(self.args))

            unlock(self.condition_truth_value)

            inversion_tracker = 1

            # This list will contain the qubits holding the truth values of
            # conditional/control environments within this environment. The instruction
            # from the subcondition environments do not need to be controlled,
            # since their compile method compiles their condition truth value based
            # on the truth value of the parent environment.
            subcondition_truth_values = [
                env.condition_truth_value for env in self.sub_condition_envs
            ]

            # Now we need to recover the instructions from the data list and
            # perform their controlled version on the condition_truth_value qubit
            while self.env_data:
                instruction = self.env_data.pop(0)

                # If the instruction == conditional environment, compile the environment
                if isinstance(instruction, (ControlEnvironment, ConditionEnvironment)):
                    instruction.compile()

                    subcondition_truth_values = [
                        env.condition_truth_value for env in self.sub_condition_envs
                    ]
                    continue

                # If the instruction == general environment, compile the instruction and
                # add the compilation result to the list of instructions that need to be
                # conditionally executed
                elif issubclass(instruction.__class__, QuantumEnvironment):
                    temp_data_list = list(self.env_qs.data)
                    self.env_qs.clear_data()
                    instruction.compile()
                    self.env_data = list(self.env_qs.data) + self.env_data
                    self.env_qs.clear_data()
                    self.env_qs.data.extend(temp_data_list)

                    subcondition_truth_values = [
                        env.condition_truth_value for env in self.sub_condition_envs
                    ]
                    continue

                if instruction.op.name in ["qb_alloc", "qb_dealloc"]:
                    self.env_qs.append(instruction)
                    continue

                if set(instruction.qubits).intersection(subcondition_truth_values):
                    self.env_qs.append(instruction)
                    continue

                # Support for inversion of a condition without opening a new environment
                if set(instruction.qubits).issubset([self.condition_truth_value]):
                    if instruction.op.name == "x":
                        inversion_tracker *= -1
                        perm_unlock(self.condition_truth_value)
                        x(self.condition_truth_value)
                        perm_lock(self.condition_truth_value)
                    elif instruction.op.name == "p":
                        p(instruction.op.params[0], self.condition_truth_value)
                    elif instruction.op.name == "rz":
                        rz(instruction.op.params[0], self.condition_truth_value)
                    elif instruction.op.name == "z":
                        z(self.condition_truth_value)
                    else:
                        raise Exception(
                            "Tried to perform invalid operations"
                            "on condition truth value (allowed are x, p, z, rz)"
                        )
                    continue

                # Create controlled instruction
                instruction.op = instruction.op.control(1)

                # Add condition truth value qubit to the instruction qubit list
                instruction.qubits = [self.condition_truth_value] + list(
                    instruction.qubits
                )
                # Append instruction
                self.env_qs.append(instruction)

            unlock(self.condition_truth_value)
            perm_unlock(self.condition_truth_value)
            if inversion_tracker == -1:
                x(self.condition_truth_value)

            perm_unlock(recursive_qv_search(self.args))

            # Uncompute truth values

            # If we had a parent environment, we first uncompute the
            # "actual truth value", i.e. the mcx of the parent and
            # this environments truth value
            if self.parent_cond_env is not None:
                self.parent_cond_env.sub_condition_envs.extend(
                    self.sub_condition_envs + [self]
                )
                mcx(toffoli_qb_list, self.qbool, method="gray_pt_inv")
                self.qbool.delete()

            # We now uncompute the environments' truth value

            # For this we can use the uncompute method, which will however not
            # recompute any intermediate values, therefore blocking alot of qubits
            # during execution. Especially in nested environments this can quickly
            # become a problem, because the blocked ancillae can not be reused
            # for further condition evaluations.

            # For the condition environment examples we saw an increase from 36 to 44
            # qubits without recomputation while only lowering the depth by about 25%

            # There might be cases where recomputation based uncomputation
            # is not worth it but for now, we leave it

            recompute = True
            # if not recompute:
            #     try:
            #         cond_eval_bool.uncompute()
            #     except:
            #         recompute = True

            if recompute:
                with invert():
                    redirected_qfunction = redirect_qfunction(self.cond_eval_function)
                    redirected_qfunction(
                        *self.args, target=cond_eval_bool, **self.kwargs
                    )

                if isinstance(self.env_qs.data[-1], QuantumEnvironment):
                    env = self.env_qs.data.pop(-1)
                    env.compile()
                
                cond_eval_bool.delete()


# This decorator allows to have conditional evaluations to return condition environments
# when called after a "with" statement but QuantumBools otherwise.

# from qrisp import QuantumFloat
# a = QuantumFloat(2)
# b = QuantumFloat(2)

# temp = (a == b) # this is a QuantumBool

# with temp: # QuantumBools have an enter method,
# but they use the less efficient control environment
#     pass

# with a == b: # This is a ConditionEnvironment entered
#     pass


# We do this mainly for performance reasons. If a QuantumBool is returned,
# it can also be entered, but it uses the less efficient (almost a factor 2)
# control Environment.

# The detection is based on the inspect module.
# This implies that the detection only works if the with statements is exactly
# two levels above the cond_eval_function call.
# That means, a class like QuantumFloat can call __eq__ and then the result
# of this decorator in order for this to work


def adaptive_condition(cond_eval_function):
    def new_cond_eval_function(*args, **kwargs):
        from qrisp import auto_uncompute

        calling_line = find_calling_line(2)

        uncomputed_function = auto_uncompute(cond_eval_function)

        if calling_line.split(" ")[0] == "with" and "&" not in calling_line and "|" not in calling_line and "~" not in calling_line:
            return quantum_condition(uncomputed_function)(*args, **kwargs)
        else:
            return uncomputed_function(*args, **kwargs)

    return new_cond_eval_function


@adaptive_condition
def q_eq(input_0, input_1):
    from qrisp import cx
    from qrisp.qtypes.quantum_bool import QuantumBool

    res = QuantumBool(name="eq_cond*")

    if isinstance(input_1, QuantumVariable):
        if input_0.size != input_1.size:
            raise Exception(
                "Tried to evaluate equality conditional"
                "for QuantumVariables of differing size"
            )

        temp_qv = QuantumVariable(input_0.size)
        cx(input_0, temp_qv)
        cx(input_1, temp_qv)
        x(temp_qv)
        mcx(temp_qv, res, method="gray_pt")

        return res

    else:
        label_int = input_0.encoder(input_1)

        mcx(input_0, res, ctrl_state=label_int)

        return res



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.circuit import QubitAlloc, QubitDealloc
from qrisp.environments.quantum_environments import QuantumEnvironment


# Environment inheritor where the environment content is appended as the inverse
class InversionEnvironment(QuantumEnvironment):
    """
    This QuantumEnvironment can be used to invert (i.e. "dagger") a block of operations.

    An alias for this is ``invert``.

    Examples
    --------

    We increment a :ref:`QuantumFloat` and afterwards revert using the
    InversionEnvironment: ::

        from qrisp import QuantumFloat, invert

        qf = QuantumFloat(4)

        qf += 3

        with invert():
            qf += 3


    >>> print(qf)
    {0: 1.0}
    >>> print(qf.qs)
    QuantumCircuit:
    ---------------
          ┌───────────┐┌──────────────┐
    qf.0: ┤0          ├┤0             ├
          │           ││              │
    qf.1: ┤1          ├┤1             ├
          │  __iadd__ ││  __iadd___dg │
    qf.2: ┤2          ├┤2             ├
          │           ││              │
    qf.3: ┤3          ├┤3             ├
          └───────────┘└──────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumFloat qf

    In the next example, we create a :ref:`QuantumFloat` and bring it into uniform
    superposition. We calculate the square and set a :ref:`QuantumBool` to ``True``,
    based on if the result is less than 10. Finally, we use the InversionEnvironment
    to uncompute the result of the multiplication. ::

        from qrisp import QuantumBool, h, q_mult, multi_measurement

        qf = QuantumFloat(3)

        h(qf)

        mult_res = q_mult(qf, qf)

        q_bool = QuantumBool()

        with mult_res < 10:
            q_bool.flip()


        with invert():
            q_mult(qf, qf, target = mult_res)


        mult_res.delete(verify = True)

    >>> print(multi_measurement([qf, q_bool]))
    {(0, True): 0.125, (1, True): 0.125, (2, True): 0.125, (3, True): 0.125,
    (4, False): 0.125, (5, False): 0.125, (6, False): 0.125, (7, False): 0.125}

    .. note::
        In many cases, this way of manually uncomputing only works if the uncomputed
        function (in this case ``q_mult``) allows specifying the target variable.
        Using the :meth:`redirect_qfunction <qrisp.redirect_qfunction>` decorator, you
        can turn any quantum function into it's target specifiable version.

    """

    def __enter__(self):
        self.manual_allocation_management = True

        QuantumEnvironment.__enter__(self)

    def compile(self):
        # Save original circuit
        original_circuit = self.env_qs.copy()

        self.env_qs.clear_data()

        # Compile environment
        for instruction in self.env_data:
            # If the instruction is an environment, compile the environment
            if isinstance(instruction, QuantumEnvironment):
                instruction.compile()
                continue

            self.env_qs.append(instruction)

        # We are now faced with the challenge of handling the (De)Allocation gates.
        # Allocation gates are inverted to Deallocation and vice versa.
        # Implying no treatment at all can lead to the situation that a QuantumVariable
        # created inside this environment is unintentionally deallocated after
        # compilation and has no prior allocation.

        # from qrisp import invert, QuantumBool, QuantumFloat, cx

        # qf = QuantumFloat(3)

        # with invert():
        #     qf_res = qf*qf

        # print(qf.qs)

        # QuantumCircuit:
        # ---------------
        #              ┌──────────┐┌──────────────┐
        #        qf.0: ┤ qb_alloc ├┤0             ├──────────────
        #              ├──────────┤│              │
        #        qf.1: ┤ qb_alloc ├┤1             ├──────────────
        #              ├──────────┤│              │
        #        qf.2: ┤ qb_alloc ├┤2             ├──────────────
        #              └──────────┘│              │┌────────────┐
        # mul_res_1.0: ────────────┤3             ├┤ qb_dealloc ├
        #                          │              │├────────────┤
        # mul_res_1.1: ────────────┤4             ├┤ qb_dealloc ├
        #                          │              │├────────────┤
        # mul_res_1.2: ────────────┤5             ├┤ qb_dealloc ├
        #                          │   __mul___dg │├────────────┤
        # mul_res_1.3: ────────────┤6             ├┤ qb_dealloc ├
        #                          │              │├────────────┤
        # mul_res_1.4: ────────────┤7             ├┤ qb_dealloc ├
        #                          │              │├────────────┤
        # mul_res_1.5: ────────────┤8             ├┤ qb_dealloc ├
        #              ┌──────────┐│              │├────────────┤
        # sbp_anc_3.0: ┤ qb_alloc ├┤9             ├┤ qb_dealloc ├
        #              ├──────────┤│              │├────────────┤
        # sbp_anc_4.0: ┤ qb_alloc ├┤10            ├┤ qb_dealloc ├
        #              ├──────────┤│              │├────────────┤
        # sbp_anc_5.0: ┤ qb_alloc ├┤11            ├┤ qb_dealloc ├
        #              └──────────┘└──────────────┘└────────────┘
        # Live QuantumVariables:
        # ----------------------
        # QuantumFloat mul_res_1
        # QuantumFloat qf

        # A naive solution would be to collect all (De)allocation gates and execute them
        # after and before the inverted instructions are appended.

        # This however results in the problem that automatic recomputation ancilla
        # management is no longer available within this environment because a function
        # using ancillae being uncomputed, will recompute it's ancillae on the same
        # qubits because there are no longer any (De)Allocation gates in between.

        # from qrisp import invert, QuantumBool, QuantumFloat, cx

        # qf = QuantumFloat(3)

        # with invert():
        #     qf_res = qf*qf
        #     qb = QuantumBool()
        #     cx(qf_res[0], qb)
        #     qf_res.uncompute()

        # print(qf.qs)

        # QuantumCircuit:
        # ---------------
        #              ┌──────────┐┌─────────────────┐     ┌──────────────┐
        #        qf.0: ┤ qb_alloc ├┤0                ├─────┤0             ├──────────────
        #              ├──────────┤│                 │     │              │
        #        qf.1: ┤ qb_alloc ├┤1                ├─────┤1             ├──────────────
        #              ├──────────┤│                 │     │              │
        #        qf.2: ┤ qb_alloc ├┤2                ├─────┤2             ├──────────────
        #              ├──────────┤│                 │     │              │┌────────────┐
        # mul_res_1.0: ┤ qb_alloc ├┤3                ├──■──┤3             ├┤ qb_dealloc ├
        #              ├──────────┤│                 │  │  │              │├────────────┤
        # mul_res_1.1: ┤ qb_alloc ├┤4                ├──┼──┤4             ├┤ qb_dealloc ├
        #              ├──────────┤│                 │  │  │              │├────────────┤
        # mul_res_1.2: ┤ qb_alloc ├┤5                ├──┼──┤5             ├┤ qb_dealloc ├
        #              ├──────────┤│   __mul___dg_dg │  │  │   __mul___dg │├────────────┤
        # mul_res_1.3: ┤ qb_alloc ├┤6                ├──┼──┤6             ├┤ qb_dealloc ├
        #              ├──────────┤│                 │  │  │              │├────────────┤
        # mul_res_1.4: ┤ qb_alloc ├┤7                ├──┼──┤7             ├┤ qb_dealloc ├
        #              ├──────────┤│                 │  │  │              │├────────────┤
        # mul_res_1.5: ┤ qb_alloc ├┤8                ├──┼──┤8             ├┤ qb_dealloc ├
        #              ├──────────┤│                 │  │  │              │├────────────┤
        # sbp_anc_3.0: ┤ qb_alloc ├┤9                ├──┼──┤9             ├┤ qb_dealloc ├
        #              ├──────────┤│                 │  │  │              │├────────────┤
        # sbp_anc_4.0: ┤ qb_alloc ├┤10               ├──┼──┤10            ├┤ qb_dealloc ├
        #              ├──────────┤│                 │  │  │              │├────────────┤
        # sbp_anc_5.0: ┤ qb_alloc ├┤11               ├──┼──┤11            ├┤ qb_dealloc ├
        #              ├──────────┤└─────────────────┘┌─┴─┐└──────────────┘└────────────┘
        #        qb.0: ┤ qb_alloc ├───────────────────┤ X ├──────────────────────────────
        #              └──────────┘                   └───┘
        # Live QuantumVariables:
        # ----------------------
        # QuantumBool qb
        # QuantumFloat qf

        # Our solution is therefore to collect all the INITAL and FINAL allocation gates
        # and execute them separately while keeping the inner (De)Allocation gates
        # where they are.

        # from qrisp import invert, QuantumBool, QuantumFloat, cx

        # qf = QuantumFloat(3)

        # with invert():
        #     qf_res = qf*qf
        #     qb = QuantumBool()
        #     cx(qf_res[0], qb)
        #     qf_res.uncompute()

        # print(qf.qs)

        # QuantumCircuit:
        # ---------------
        #              ┌──────────┐┌─────────────────┐                               »
        #        qf.0: ┤ qb_alloc ├┤0                ├───────────────────────────────»
        #              ├──────────┤│                 │                               »
        #        qf.1: ┤ qb_alloc ├┤1                ├───────────────────────────────»
        #              ├──────────┤│                 │                               »
        #        qf.2: ┤ qb_alloc ├┤2                ├───────────────────────────────»
        #              ├──────────┤│                 │                               »
        # mul_res_2.0: ┤ qb_alloc ├┤3                ├────────────────■──────────────»
        #              ├──────────┤│                 │                │              »
        # mul_res_2.1: ┤ qb_alloc ├┤4                ├────────────────┼──────────────»
        #              ├──────────┤│                 │                │              »
        # mul_res_2.2: ┤ qb_alloc ├┤5                ├────────────────┼──────────────»
        #              ├──────────┤│   __mul___dg_dg │                │              »
        # mul_res_2.3: ┤ qb_alloc ├┤6                ├────────────────┼──────────────»
        #              ├──────────┤│                 │                │              »
        # mul_res_2.4: ┤ qb_alloc ├┤7                ├────────────────┼──────────────»
        #              ├──────────┤│                 │                │              »
        # mul_res_2.5: ┤ qb_alloc ├┤8                ├────────────────┼──────────────»
        #              ├──────────┤│                 │┌────────────┐  │  ┌──────────┐»
        # sbp_anc_6.0: ┤ qb_alloc ├┤9                ├┤ qb_dealloc ├──┼──┤ qb_alloc ├»
        #              ├──────────┤│                 │├────────────┤  │  ├──────────┤»
        # sbp_anc_7.0: ┤ qb_alloc ├┤10               ├┤ qb_dealloc ├──┼──┤ qb_alloc ├»
        #              ├──────────┤│                 │├────────────┤  │  ├──────────┤»
        # sbp_anc_8.0: ┤ qb_alloc ├┤11               ├┤ qb_dealloc ├──┼──┤ qb_alloc ├»
        #              ├──────────┤└─────────────────┘└────────────┘┌─┴─┐└──────────┘»
        #        qb.0: ┤ qb_alloc ├─────────────────────────────────┤ X ├────────────»
        #              └──────────┘                                 └───┘            »
        # «             ┌──────────────┐
        # «       qf.0: ┤0             ├──────────────
        # «             │              │
        # «       qf.1: ┤1             ├──────────────
        # «             │              │
        # «       qf.2: ┤2             ├──────────────
        # «             │              │┌────────────┐
        # «mul_res_2.0: ┤3             ├┤ qb_dealloc ├
        # «             │              │├────────────┤
        # «mul_res_2.1: ┤4             ├┤ qb_dealloc ├
        # «             │              │├────────────┤
        # «mul_res_2.2: ┤5             ├┤ qb_dealloc ├
        # «             │   __mul___dg │├────────────┤
        # «mul_res_2.3: ┤6             ├┤ qb_dealloc ├
        # «             │              │├────────────┤
        # «mul_res_2.4: ┤7             ├┤ qb_dealloc ├
        # «             │              │├────────────┤
        # «mul_res_2.5: ┤8             ├┤ qb_dealloc ├
        # «             │              │├────────────┤
        # «sbp_anc_6.0: ┤9             ├┤ qb_dealloc ├
        # «             │              │├────────────┤
        # «sbp_anc_7.0: ┤10            ├┤ qb_dealloc ├
        # «             │              │├────────────┤
        # «sbp_anc_8.0: ┤11            ├┤ qb_dealloc ├
        # «             └──────────────┘└────────────┘
        # «       qb.0: ──────────────────────────────
        # «
        # Live QuantumVariables:
        # ----------------------
        # QuantumBool qb
        # QuantumFloat qf

        initially_allocated_qubits = []
        i = 0
        while i < len(self.env_qs.data):
            instr = self.env_qs.data[i]
            if (
                instr.op.name == "qb_alloc"
                and not instr.qubits[0] in initially_allocated_qubits
            ):
                initially_allocated_qubits.append(self.env_qs.data.pop(i).qubits[0])
                continue
            i += 1

        deallocated_qubits = []
        i = 0

        self.env_qs.data.reverse()
        while i < len(self.env_qs.data):
            instr = self.env_qs.data[i]
            if (
                instr.op.name == "qb_dealloc"
                and not instr.qubits[0] in deallocated_qubits
            ):
                deallocated_qubits.append(self.env_qs.data.pop(i).qubits[0])
                continue
            i += 1

        deallocated_qubits = list(set(deallocated_qubits))

        self.env_qs.data.reverse()

        # print(transpile(self.env_qs.inverse()))
        # print(transpile(self.env_qs))
        # Merge the original circuit with the inverse of the environment content

        original_circuit.qubits = self.env_qs.qubits
        original_circuit.clbits = self.env_qs.clbits

        for qubit in initially_allocated_qubits:
            original_circuit.append(QubitAlloc(), [qubit])

        original_circuit.extend(self.env_qs.inverse())

        for qubit in deallocated_qubits:
            original_circuit.append(QubitDealloc(), [qubit])

        # Reinstate the resulting circuit in the quantum session circuit
        self.env_qs.data = original_circuit.data


# Shortcut to quickly initiate inversion environments
def invert():
    return InversionEnvironment()



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.interface import BackendServer, BackendClient
from qrisp import QuantumVariable, QuantumCircuit, h, x, cx


def test_interface():
    def run_func(qc, shots, token):
        print("Recieved query")

        # Convert to qiskit
        from qrisp.interface.circuit_converter import convert_circuit

        qiskit_qc = convert_circuit(qc, "qiskit")

        print(qiskit_qc)

        from qiskit import Aer

        qiskit_backend = Aer.get_backend("qasm_simulator")

        # Run Circuit on the Qiskit backend
        return qiskit_backend.run(qiskit_qc, shots=shots).result().get_counts()

    #socket_ip = "::"
    socket_ip = "127.0.0.1"
    port = 8083

    test_server = BackendServer(run_func, socket_ip_address=socket_ip, port=port)

    test_server.start()

    qv1 = QuantumVariable(2)
    qv2 = QuantumVariable(2)

    h(qv1[0])
    x(qv1[1])
    cx(qv1[0], qv2[1])

    test_client = BackendClient(socket_ip, port)

    res = qv1.get_measurement(backend=test_client)
    assert "01" in res.keys() and "11" in res.keys()



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.interface.circuit_converter import convert_circuit
from qrisp.interface.openapi_interface.codegen.client.openapi_client import (
    ApiClient,
    Configuration,
)
from qrisp.interface.openapi_interface.codegen.client.openapi_client.api.default_api \
    import DefaultApi
from qrisp.interface.openapi_interface.codegen.client.openapi_client.models import (
    InlineObject,
)


class BackendClient(DefaultApi):
    def __init__(self, socket_ip, port):
        if socket_ip.find(":") != -1:
            socket_ip = "[" + socket_ip + "]"

        if port is None:
            port = 9010

        # TO-DO Allow API token/Secure connections ...
        # check Configuration class definition
        config = Configuration(host="http://" + socket_ip + ":" + str(port))
        client = ApiClient(configuration=config)

        super().__init__(client)

    def run(self, qc, shots, token=""):
        request_object = InlineObject(
            qc=convert_circuit(qc, "open_api"), shots=shots, token=token
        )
        return super().run(inline_object=request_object)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.environments.quantum_environments import QuantumEnvironment
from qrisp.misc.GMS_tools import GXX_converter


# Environments that allows quick and easy access to the GMS_converter
# To use it make sure every gate that happens inside is either phase or cphase
class GMSEnvironment(QuantumEnvironment):
    """
    This environment provides a convenient interface for constructing quantum algorithms
    using the Ion-trap native GMS gates. GMS gates allow entangling more than 2 qubits
    in a single step and can therefore boost performance in many situations. For more
    information on GMS gates consult https://arxiv.org/abs/quant-ph/9810040 . The
    techniques for converting the circuits presented to this environment are mostly
    based on https://ieeexplore.ieee.org/document/9815035 .

    This environment allows to code blocks of phase-only gates as we are used to but
    compiles these blocks to GMS gates.

    Examples
    --------


    We create a function performing the quantum Fourier-transform using GMS gates: ::

        from qrisp import QuantumEnvironment, GMSEnvironment, h, cp, swap
        import numpy as np

        def QFT(qv, use_gms = False):

            n = qv.size

            if use_gms:
                env = GMSEnvironment
            else:
                env = QuantumEnvironment


            for i in range(n):
                h(qv[i])

                if i == n-1:
                    break


                with env():

                    #This is the block which converted to GMS gates
                    #We can only use the gates p, cp and rz in here

                    for k in range(n-i-1):
                        cp(2*np.pi/2**(k+2), qv[k+i+1], qv[i])

            for i in range(n//2):
                swap(qv[i], qv[n-i-1])


    We inspect the resulting quantum circuit:

    >>> from qrisp import QuantumFloat, invert
    >>> qf = QuantumFloat(5)
    >>> qf[:] = 13
    >>> QFT(qf, use_gms = True)
    >>> print(qf.qs)
    QuantumCircuit:
    ---------------
          ┌───┐┌───┐┌─────────────────────┐                                 »
    qf.0: ┤ X ├┤ H ├┤0                    ├─────────────────────────────────»
          └───┘└───┘│                     │┌───┐┌─────────────────────┐     »
    qf.1: ──────────┤1                    ├┤ H ├┤0                    ├─────»
          ┌───┐     │                     │└───┘│                     │┌───┐»
    qf.2: ┤ X ├─────┤2 GXX converted gate ├─────┤1                    ├┤ H ├»
          ├───┤     │                     │     │  GXX converted gate │└───┘»
    qf.3: ┤ X ├─────┤3                    ├─────┤2                    ├─────»
          └───┘     │                     │     │                     │     »
    qf.4: ──────────┤4                    ├─────┤3                    ├─────»
                    └─────────────────────┘     └─────────────────────┘     »
    «
    «qf.0: ─────────────────────────────────────────────────────────X─
    «                                                               │
    «qf.1: ─────────────────────────────────────────────────────X───┼─
    «      ┌─────────────────────┐                              │   │
    «qf.2: ┤0                    ├──────────────────────────────┼───┼─
    «      │                     │┌───┐┌─────────────────────┐  │   │
    «qf.3: ┤1 GXX converted gate ├┤ H ├┤0                    ├──X───┼─
    «      │                     │└───┘│  GXX converted gate │┌───┐ │
    «qf.4: ┤2                    ├─────┤1                    ├┤ H ├─X─
    «      └─────────────────────┘     └─────────────────────┘└───┘
    Live QuantumVariables:
    ----------------------
    QuantumFloat qf

    Now we check that the GMS version indeed performs the same operation as the CNOT
    version by performing the inverse of the CNOT version.

    >>> with invert(): QFT(qf, use_gms = False)
    >>> print(qf)
    {13: 1.0}

    """

    # We only need to modify the compile method of the base environment class
    def compile(self):
        # Temporarily store the data of the quantum session
        temp_data = list(self.env_qs.data)
        self.env_qs.clear_data()

        QuantumEnvironment.compile(self)

        copied_circ = self.env_qs.copy()
        depth_dic = self.env_qs.get_depth_dic()

        i = 0
        while i < len(copied_circ.qubits):
            if depth_dic[copied_circ.qubits[i]]:
                i += 1
            else:
                copied_circ.qubits.pop(i)

        if not len(copied_circ.qubits):
            return

        # Convert the circuit
        # converted_gate, qubit_map = GMS_converter(self.env_qs, True)
        converted_gate = GXX_converter(copied_circ).to_gate()

        converted_gate.name = "GXX converted gate"

        self.env_qs.clear_data()
        # Recover original circuit
        self.env_qs.data.extend(temp_data)

        # Apply original circuit
        self.env_qs.append(converted_gate, copied_circ.qubits)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np

import qrisp.circuit.standard_operations as std_ops
from qrisp.core import recursive_qs_search

def append_operation(operation, qubits=[], clbits=[]):
    
    qs_list = recursive_qs_search([qubits, clbits])

    if len(qs_list) == 0:
        raise Exception("Could not find QuantumSession object to append operation")

    qs_list[0].append(operation, qubits, clbits)


def cx(control, target):
    """
    Applies a CX gate.

    Parameters
    ----------
    control : Qubit or list[Qubit] or QuantumVariable
        The Qubit to control on.
    target : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the X gate on.
    """
    append_operation(std_ops.CXGate(), [control, target])

    # std_ops.CXGate().append([qubits_0, qubits_1])

    return control, target


def cy(control, target):
    """
    Applies a CY gate.

    Parameters
    ----------
    control : Qubit or list[Qubit] or QuantumVariable
        The Qubit to control on.
    target : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the Y gate on.
    """

    append_operation(std_ops.CYGate(), [control, target])
    return control, target


def cz(control, target):
    """
    Applies a CZ gate.

    Parameters
    ----------
    control : Qubit or list[Qubit] or QuantumVariable
        The Qubit to control on.
    target : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the Z gate on.
    """

    append_operation(std_ops.CZGate(), [control, target])
    return control, target


def h(qubits):
    """
    Applies an H gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the H gate on.
    """
    append_operation(std_ops.HGate(), [qubits])

    return qubits


def x(qubits):
    """
    Applies an X gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the X gate on.
    """

    append_operation(std_ops.XGate(), [qubits])

    return qubits


def y(qubits):
    """
    Applies a Y gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the Y gate on.
    """

    append_operation(std_ops.YGate(), [qubits])

    return qubits


def z(qubits):
    """
    Applies a Z gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the Z gate on.
    """
    append_operation(std_ops.ZGate(), [qubits])

    return qubits


def mcx(controls, target, method="auto", ctrl_state=-1, num_ancilla=1):
    r"""
    Applies a multi-controlled X gate.

    The following methods are available:

    * ``gray`` performs a gray code traversal which requires no ancillae 
      but is rather inefficient for large numbers of control qubits.
    
    * ``gray_pt`` and ``gray_pt_inv`` are more efficient but introduce extra 
      phases that need to be uncomputed by performing the inverse of this gate 
      on the same inputs. For more information on phase tolerance, check
      `this paper <https://iopscience.iop.org/article/10.1088/2058-9565/acaf9d/meta>`_.
    
    * ``balauca`` is a method based on this `paper
      <https://www.iccs-meeting.org/archive/iccs2022/papers/133530169.pdf>`_ with
      logarithmic depth but requires many ancilla qubits.
    
    * ``maslov`` is documented `here <https://arxiv.org/abs/1508.03273>`_, requires less
      ancilla qubits but is only available for 4 or less control qubits.
    
    * ``yong`` can be found int this `article
      <https://link.springer.com/article/10.1007/s10773-017-3389-4>`_.This method requires
      only a single ancilla and has moderate scaling in depth and gate count.
    
    * ``hybrid`` is a flexible method which combines the other available methods, such
      that the amount of used ancillae is customizable. After several ``balauca``-layers,
      the recursion is canceled by either a ``yong``, ``maslov`` or ``gray`` mcx,
      depending on what fits the most.
    
    * ``auto`` recompiles the mcx gate at compile time using the hybrid algorithm
      together with the information about how many clean/dirty ancillae qubits are
      available. For more information check :meth:`qrisp.QuantumSession.compile`.
    
    
    .. note::
        Due to Qrisp's automatic qubit management, clean ancilla qubits are not as much
        of a sparse resource as one might think. Even though the ``balauca`` method
        requires a considerable amount of ancillae, many other functions also do,
        implying there is alot of recycling potential. The net effect is that in more
        complex programs, the amount of qubits of the circuit returned by the
        :meth:`compile method <qrisp.QuantumSession.compile>` increases only slightly.


    Parameters
    ----------
    controls : list[Qubits] or QuantumVariable
        The Qubits to control on.
    target : Qubit
        The Qubit to perform the X gate on.
    method : str, optional
        The synthesis method. Available are ``auto``, ``gray``, ``gray_pt``,
        ``gray_pt_inv``, ``maslov``, ``balauca`` and ``yong``. The default is ``auto``.
    ctrl_state : int or str, optional
        The state on which to activate the X gate. The default is "1111..".
    num_ancilla : int, optional
        Specifies the amount of ancilla qubits to use. This parameter is used only if
        the method is set to ``hybrid``. The default is 1.


    Examples
    --------

    We apply a 3-contolled X gate

    >>> from qrisp import QuantumVariable, mcx
    >>> control = QuantumVariable(3)
    >>> target = QuantumVariable(1)
    >>> mcx(control, target, method = "gray")
    >>> print(control.qs)
    QuantumCircuit:
    ---------------
    control.0: ──■──
                 │
    control.1: ──■──
                 │
    control.2: ──■──
               ┌─┴─┐
     target.0: ┤ X ├
               └───┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable control
    QuantumVariable target

    We compare different performance indicators. ::

        from qrisp import QuantumVariable, mcx

        def benchmark_mcx(n, methods):
            for method in methods:

                controls = QuantumVariable(n)
                target = QuantumVariable(1)

                mcx(controls, target, method = method)

                compiled_qc = controls.qs.compile()

                print(f"==================\nMethod: {method}\n------------------")
                print(f"Depth: \t\t\t{compiled_qc.depth()}")
                print(f"CNOT count: \t{compiled_qc.cnot_count()}")
                print(f"Qubit count: \t{compiled_qc.num_qubits()}")


    >>> benchmark_mcx(4, methods = ["gray", "gray_pt", "maslov", "balauca", "yong"])
    ==================
    Method: gray
    ------------------
    Depth:          50
    CNOT count:     30
    Qubit count:    5
    ==================
    Method: gray_pt
    ------------------
    Depth:          34
    CNOT count:     16
    Qubit count:    5
    ==================
    Method: maslov
    ------------------
    Depth:          43
    CNOT count:     18
    Qubit count:    6
    ==================
    Method: balauca
    ------------------
    Depth:          22
    CNOT count:     18
    Qubit count:    7
    ==================
    Method: yong
    ------------------
    Depth:          77
    CNOT count:     30
    Qubit count:    6
    >>> benchmark_mcx(10, methods = ["gray", "gray_pt", "balauca", "yong"])
    ==================
    Method: gray
    ------------------
    Depth:          3106
    CNOT count:     2046
    Qubit count:    11
    ==================
    Method: gray_pt
    ------------------
    Depth:          2050
    CNOT count:     1024
    Qubit count:    11
    ==================
    Method: balauca
    ------------------
    Depth:          53
    CNOT count:     54
    Qubit count:    18
    ==================
    Method: yong
    ------------------
    Depth:          621
    CNOT count:     264
    Qubit count:    12

    """

    from qrisp.circuit.quantum_circuit import convert_to_qb_list
    from qrisp.misc import bin_rep

    qubits_0 = convert_to_qb_list(controls)
    qubits_1 = convert_to_qb_list(target)

    n = len(qubits_0)

    if n == 0:
        return controls, target

    if not isinstance(ctrl_state, str):
        if ctrl_state == -1:
            ctrl_state += 2**n
        ctrl_state = bin_rep(ctrl_state, n)[::-1]

    if len(ctrl_state) != n:
        raise Exception(
            f"Given control state {ctrl_state} does not match control qubit amount {n}"
        )

    from qrisp.misc.multi_cx import (
        balauca_dirty,
        balauca_mcx,
        hybrid_mcx,
        maslov_mcx,
        yong_mcx,
    )

    if method in ["gray", "gray_pt", "gray_pt_inv"] or len(qubits_0) == 1:
        if len(qubits_0) == 1:
            method = "gray"
        append_operation(
            std_ops.MCXGate(len(qubits_0), ctrl_state, method=method),
            qubits_0 + qubits_1,
        )

    elif method == "maslov":
        from qrisp import QuantumBool

        if n >= 3:
            ancilla = [QuantumBool(name="maslov_anc_")]
        else:
            ancilla = []
        append_operation(maslov_mcx(n, ctrl_state), qubits_0 + ancilla + qubits_1)

        [qv.delete() for qv in ancilla]

    elif method == "balauca":
        balauca_mcx(qubits_0, qubits_1, ctrl_state=ctrl_state)

    elif method == "balauca_dirty":
        balauca_dirty(qubits_0, qubits_1, k=num_ancilla, ctrl_state=ctrl_state)

    elif method == "yong":
        yong_mcx(qubits_0, qubits_1, ctrl_state=ctrl_state)

    elif method == "hybrid":
        hybrid_mcx(qubits_0, qubits_1, ctrl_state=ctrl_state, num_ancilla=num_ancilla)

    elif method == "auto":
        # if n <= 3:
        #     return mcx(qubits_0, qubits_1, method = "gray", ctrl_state = ctrl_state)
        # if 3 < n < 5:
        #     return mcx(qubits_0, qubits_1, method = "maslov", ctrl_state = ctrl_state)
        # else:
        #     return mcx(qubits_0, qubits_1, method = "balauca", ctrl_state = ctrl_state) # noqa:501

        gate = std_ops.MCXGate(len(qubits_0), ctrl_state, method="auto")
        append_operation(gate, qubits_0 + qubits_1)

    return controls, target


def mcz(qubits, method="auto", ctrl_state=-1, num_ancilla=1):
    """
    Applies a multi-controlled Z gate.

    For more information on the available methods, check
    :meth:`the mcx documentation page <qrisp.mcx>`.

    Parameters
    ----------
    qubits : QuantumVariable or list[Qubits]
        The Qubits to control on.
    method : str, optional
        The synthesis method. Available are ``auto``, ``gray``, ``gray_pt``,
        ``gray_pt_inv``, ``maslov``, ``balauca``, ``yong`` and ``hybrid``. The default
        is ``auto``.
    ctrl_state : int or str, optional
        The state on which to activate the Z gate. The default is "1111...".
    num_ancilla : int, optional
        Specifies the amount of ancilla qubits to use. This parameter is used only if
        the method is set to ``hybrid``. The default is 1.

    """

    from qrisp.misc import gate_wrap

    @gate_wrap(permeability="full", is_qfree=True, name="anc supported mcz")
    def mcz_inner(qubits, method="auto", ctrl_state=-1):
        if len(ctrl_state) != n:
            raise Exception(
                f"Given control state {ctrl_state} does not match"
                f"control qubit amount {n}"
            )

        from qrisp import h, x

        if ctrl_state[-1] == "0":
            x(qubits[-1])

        h(qubits[-1])
        mcx(qubits[:-1], qubits[-1], method=method, ctrl_state=ctrl_state[:-1])
        h(qubits[-1])

        if ctrl_state[-1] == "0":
            x(qubits[-1])

        return qubits

    n = len(qubits)

    from qrisp import bin_rep

    if not isinstance(ctrl_state, str):
        if ctrl_state == -1:
            ctrl_state += 2**n
        ctrl_state = bin_rep(ctrl_state, n)

    if method in ["gray", "auto"]:
        if ctrl_state[-1] == "0":
            x(qubits[-1])

        append_operation(
            std_ops.ZGate().control(
                len(qubits) - 1, method=method, ctrl_state=ctrl_state[:-1]
            ),
            qubits,
        )

        if ctrl_state[-1] == "0":
            x(qubits[-1])

        return qubits

    return mcz_inner(qubits, method, ctrl_state)


def mcp(phi, qubits, method="auto", ctrl_state=-1):
    """
    Applies a multi-controlled phase gate.

    The available methods are:

    * ``gray`` , which performs a traversal of the gray code.

    * ``balauca`` , which is a modified version of the algorithm presented `here
      <https://www.iccs-meeting.org/archive/iccs2022/papers/133530169.pdf>`_.

    * ``auto`` , which picks ``gray`` for any qubit count less than 4 and ``balauca``
      otherwise.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The phase to apply.
    qubits : list[Qubit] or QuantumVariable
        The qubits to apply the multi-controlled phase gate on.
    method : str, optional
        The method to deploy. The default is "auto".
    ctrl_state : str or int, optional
        The control state on which to apply the phase. The default is "111...".

    """

    from qrisp.misc.multi_cx import balauca_mcx
    from qrisp.misc import bin_rep, gate_wrap

    @gate_wrap(permeability="full", is_qfree=True, name="anc supported mcp")
    def balauca_mcp(phi, qubits, ctrl_state):
        from qrisp.circuit.quantum_circuit import convert_to_qb_list

        qubits = convert_to_qb_list(qubits)
        if ctrl_state[-1] == "0":
            x(qubits[-1])

        balauca_mcx(qubits[:-1], [qubits[-1]], ctrl_state=ctrl_state, phase=phi)

        if ctrl_state[-1] == "0":
            x(qubits[-1])

    n = len(qubits)

    if not isinstance(ctrl_state, str):
        if ctrl_state == -1:
            ctrl_state += 2**n
        ctrl_state = bin_rep(ctrl_state, n)[::-1]

    n = len(qubits)

    if method == "gray":
        if ctrl_state[-1] == "0":
            x(qubits[-1])

        append_operation(
            std_ops.PGate(phi).control(n - 1, ctrl_state=ctrl_state[:-1]), qubits
        )

        if ctrl_state[-1] == "0":
            x(qubits[-1])
        return qubits

    elif method == "balauca":
        balauca_mcp(phi, qubits, ctrl_state=ctrl_state)
        return qubits

    elif method == "auto":
        if n < 4:
            return mcp(phi, qubits, method="gray", ctrl_state=ctrl_state)
        else:
            return mcp(phi, qubits, method="balauca", ctrl_state=ctrl_state)

    else:
        raise Exception(f"Don't know method {method}")


def p(phi, qubits):
    """
    Applies a phase gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The phase to apply.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit on which to apply the phase gate.

    """

    append_operation(std_ops.PGate(phi), [qubits])

    # std_ops.PGate(phi).append([qubits])

    return qubits


def cp(phi, qubits_0, qubits_1):
    """
    Applies a controlled phase gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The phase to apply.
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first Qubit.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second Qubit.

    """

    append_operation(std_ops.CPGate(phi), [qubits_0, qubits_1])

    # std_ops.CPGate(phi).append([qubits_0, qubits_1])

    return qubits_0, qubits_1


def rx(phi, qubits):
    """
    Applies an RX gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The angle parameter.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the RX gate on.

    """

    append_operation(std_ops.RXGate(phi), [qubits])

    return qubits


def ry(phi, qubits):
    """
    Applies an RY gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The angle parameter.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the RY gate on.

    """

    append_operation(std_ops.RYGate(phi), [qubits])

    return qubits


def rz(phi, qubits):
    """
    Applies an RY gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The angle parameter.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the RY gate on.

    """

    append_operation(std_ops.RZGate(phi), [qubits])

    return qubits


def crz(phi, qubits_0, qubits_1):
    """
    Applies controled RZ gate

    Parameters
    ----------
    phi : float or sympy.Symbol
        The angle parameter.
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first Qubit.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second Qubit.

    """

    append_operation(std_ops.RZGate(phi).control(1), [qubits_0, qubits_1])
    return qubits_0, qubits_1


def s(qubits):
    """
    Applies an S gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the S gate on.
    """

    append_operation(std_ops.SGate(), [qubits])
    return qubits


def t(qubits):
    """
    Applies a T gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the T gate on.
    """
    append_operation(std_ops.TGate(), [qubits])
    return qubits


def s_dg(qubits):
    """
    Applies a daggered S gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the daggered S gate on.
    """

    append_operation(std_ops.SGate().inverse(), [qubits])
    return qubits


def t_dg(phi, qubits):
    """
    Applies a daggered T gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the daggered T gate on.
    """
    append_operation(std_ops.TGate().inverse(), [qubits])
    return qubits


def sx(qubits):
    """
    Applies an SX gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the SX gate on.
    """

    append_operation(std_ops.SXGate().inverse(), [qubits])

    return qubits


def sx_dg(qubits):
    """
    Applies a daggered SX gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the daggered SX gate on.
    """
    append_operation(std_ops.SXDGGate().inverse(), [qubits])
    return qubits


def gphase(phi, qubits):
    """
    Applies a global phase. This gate turns into a phase gate when controlled.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The global phase to apply.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the global phase gate on.
    """

    append_operation(std_ops.GPhaseGate(phi), qubits)
    return qubits



def xxyy(phi, beta, qubits_0, qubits_1):
    """
    Applies an XXYY interaction gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The global phase to apply.
    beta : float or sympy.Symbol
        The other angle parameter.
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first Qubit to perform the XXYY gate on.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second Qubit to perform the XXYY gate on.
    """

    append_operation(std_ops.XXYYGate(phi, beta), [qubits_0, qubits_1])
    return qubits_0, qubits_1


def rzz(phi, qubits_0, qubits_1):
    """
    Applies an RZZ gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The phase to apply.
    beta : float or sympy.Symbol
        The other angle parameter.
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first argument to perform the RZZ gate one.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second argument to perform the RZZ gate one.
    """

    append_operation(std_ops.RZZGate(phi), [qubits_0, qubits_1])
    return qubits_0, qubits_1

def rxx(phi, qubits_0, qubits_1):
    """
    Applies an RXX gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The phase to apply.
    beta : float or sympy.Symbol
        The other angle parameter.
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first argument to perform the RXX gate one.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second argument to perform the RXX gate one.
    """

    append_operation(std_ops.RXXGate(phi), [qubits_0, qubits_1])
    return qubits_0, qubits_1

def u3(theta, phi, lam, qubits):
    """
    Applies an U3 gate.

    Parameters
    ----------
    theta : float or sympy.Symbol
        The first angle parameter.
    phi : float or sympy.Symbol
        The second angle parameter.
    lambda : float or sympy.Symbol
        The third angle parameter.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the U3 gate on.
    """

    append_operation(std_ops.RZGate(phi), [qubits])

    return qubits


def measure(qubits, clbits=None):
    """
    Performs a measurement of the specified Qubit.

    Parameters
    ----------
    qubit : Qubit or list[Qubit] or QuantumVariable
        The Qubit to measure.
    clbit : Clbit, optional
        The Clbit to store the result in. By default, a new Clbit will be created.

    """
    if clbits is None:
        clbits = []
        if hasattr(qubits, "__len__"):
            for qb in qubits:
                try:
                    clbits.append(qubits[0].qs.add_clbit())
                except AttributeError:
                    clbits.append(qubits[0].qs().add_clbit())

        else:
            clbits = qubits.qs.add_clbit()
    append_operation(std_ops.Measurement(), [qubits], [clbits])

    return qubits


def barrier(qubits):
    """
    A visual marker for structuring the QuantumCircuit.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to apply the barrier to.

    Examples
    --------

    >>> from qrisp import QuantumVariable, x, y, barrier
    >>> qv = QuantumVariable(5)
    >>> x(qv)
    >>> barrier(qv)
    >>> y(qv)
    >>> print(qv.qs)
    QuantumCircuit:
    ---------------
          ┌───┐ ░ ┌───┐
    qv.0: ┤ X ├─░─┤ Y ├
          ├───┤ ░ ├───┤
    qv.1: ┤ X ├─░─┤ Y ├
          ├───┤ ░ ├───┤
    qv.2: ┤ X ├─░─┤ Y ├
          ├───┤ ░ ├───┤
    qv.3: ┤ X ├─░─┤ Y ├
          ├───┤ ░ ├───┤
    qv.4: ┤ X ├─░─┤ Y ├
          └───┘ ░ └───┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable qv

    """

    append_operation(std_ops.Barrier(len(qubits)), qubits)

    return qubits


def swap(qubits_0, qubits_1):
    """
    Applies a SWAP gate.

    Parameters
    ----------
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first Qubit.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second Qubit.

    """
    append_operation(std_ops.SwapGate(), [qubits_0, qubits_1])

    return qubits_0, qubits_1


def id(qubits):
    """
    Applies an ID gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The qubits to perform the ID gate on.

    """
    append_operation(std_ops.IDGate(), [qubits])

    return qubits


def QFT_inner(qv, exec_swap=True, qiskit_endian=True, inplace_mult=1, use_gms=False):
    from qrisp.misc import is_inv

    qv = list(qv)
    n = len(qv)

    if qiskit_endian:
        qv = qv[::-1]

    if not use_gms:
        from qrisp.environments.quantum_environments import QuantumEnvironment

        env = QuantumEnvironment

    else:
        from qrisp.environments.GMS_environment import GMSEnvironment

        env = GMSEnvironment

    if not is_inv(inplace_mult, n):
        raise Exception(
            "Tried to perform non-invertible inplace multiplication"
            "during Fourier-Transform"
        )

    for i in range(n):
        h(qv[i])

        if i == n - 1:
            break

        with env():
            for k in range(n - i - 1):
                cp(inplace_mult * 2 * np.pi / 2 ** (k + 2), qv[k + i + 1], qv[i])

    if exec_swap:
        for i in range(n // 2):
            swap(qv[i], qv[n - i - 1])

    return qv


def QFT(
    qv, inv=False, exec_swap=True, qiskit_endian=True, inplace_mult=1, use_gms=False
):
    """
    Performs the quantum fourier transform on the input.

    Parameters
    ----------
    qv : QuantumVariable
        QuantumVariable to transform (in-place).
    inv : bool, optional
        If set to True, the inverse transform will be applied. The default is False.
    exec_swap : bool, optional
        If set to False, the swaps at the end of the transformation will be skipped.
        The default is True.
    qiskit_endian : bool, optional
        If set to False the order of bits will be reversed. The default is True.
    inplace_mult : int, optional
        Allows multiplying the QuantumVariable with an extra factor during the
        transformation. For more information check `the publication
        <https://ieeexplore.ieee.org/document/9815035>`_. The default is 1.
    use_gms : bool, optional
        If set to True, the QFT will be compiled using only GMS gates as entangling
        gates. The default is False.


    """
    from qrisp import gate_wrap, invert

    name = "QFT"
    if not exec_swap:
        name += " no swap"
    if inplace_mult != 1:
        name += " inpl mult " + str(inplace_mult)

    if inv:
        with invert():
            gate_wrap(permeability=[], is_qfree=False, name=name)(QFT_inner)(
                qv,
                exec_swap=exec_swap,
                qiskit_endian=qiskit_endian,
                inplace_mult=inplace_mult,
                use_gms=use_gms,
            )
    else:
        gate_wrap(permeability=[], is_qfree=False, name=name)(QFT_inner)(
            qv,
            exec_swap=exec_swap,
            qiskit_endian=qiskit_endian,
            inplace_mult=inplace_mult,
            use_gms=use_gms,
        )

    return qv


def QPE(
    args, U, precision=None, target=None, iter_spec=False, ctrl_method=None, kwargs={}
):
    r"""
    Evaluates the `quantum phase estimation algorithm
    <https://en.wikipedia.org/wiki/Quantum_phase_estimation_algorithm>`_.

    The unitary to estimate is expected to be given as Python function, which is called
    on ``args``.

    Parameters
    ----------
    args : list
        A list of arguments (could be QuantumVariables) which represent the state,
        the quantum phase estimation is performed on.
    U : function
        A Python function, which will receive the list ``args`` as arguments in the
        course of this algorithm.
    precision : int, optional
        The precision of the estimation. The default is None.
    target : QuantumFloat, optional
        A target QuantumFloat to perform the estimation into. The default is None.
        If given neither a precision nor a target, an Exception will be raised.
    iter_spec : bool, optional
        If set to ``True``, ``U`` will be called with the additional keyword
        ``iter = i`` where ``i`` is the amount of iterations to perform (instead of
        simply calling ``U`` for ``i`` times). The default is False.
    ctrl_method : string, optional
        Allows to specify which method should be used to generate the
        controlled U circuit. For more information check
        :meth:`.control <qrisp.Operation.control>`. The default is None.
    kwargs : dict, optional
        A dictionary of keyword arguments to pass to ``U``. The default is {}.

    Raises
    ------
    Exception
        Tried to perform quantum phase estimation without precision specification.

    Returns
    -------
    res : QuantumFloat
        The QuantumFloat containing the estimated phase as a fraction of $2 \pi$.

    Examples
    --------

    We define a function that applies two phase gates onto its input and estimate the
    applied phase. ::

        from qrisp import p, QuantumVariable, QPE, multi_measurement

        def U(qv):
            x = 0.5
            y = 0.125

            p(x*2*np.pi, qv[0])
            p(y*2*np.pi, qv[1])

        qv = QuantumVariable(2)

        h(qv)

        res = QPE(qv, U, precision = 3)

    >>> multi_measurement([qv, res])
    {('00', 0.0): 0.25,
     ('10', 0.5): 0.25,
     ('01', 0.125): 0.25,
     ('11', 0.625): 0.25}
    >>> res.qs.depth()
    66

    During the phase estimation, ``U`` is called $2^{\\text{precision}}$ times. We can
    reduce that number by abusing that we can bundle repeated calls into a single call
    with a modified phase. ::

        def U(qv, iter = None):
            x = 0.5
            y = 0.125

            p(x*2*np.pi*iter, qv[0])
            p(y*2*np.pi*iter, qv[1])

        qv = QuantumVariable(2)

        h(qv)

        res = QPE(qv, U, precision = 3, iter_spec = True)

    >>> multi_measurement([qv, res])
    {('00', 0.0): 0.25,
     ('10', 0.5): 0.25,
     ('01', 0.125): 0.25,
     ('11', 0.625): 0.25}
    >>> res.qs.depth()
    34

    """

    from qrisp import QuantumFloat, control

    if target is None:
        if precision is None:
            raise Exception(
                "Tried to perform quantum phase estimation without"
                "precision specification"
            )
        qpe_res = QuantumFloat(precision, -precision, signed=False)
    else:
        qpe_res = target

    h(qpe_res)

    for i in range(qpe_res.size):
        if iter_spec:
            with control(qpe_res[i], ctrl_method=ctrl_method):
                U(args, iter=2**i, **kwargs)
        else:
            with control(qpe_res[i], ctrl_method=ctrl_method):
                for j in range(2**i):
                    U(args, **kwargs)

    QFT(qpe_res, inv=True)

    return qpe_res


def quantum_counting(qv, oracle, precision):
    """
    This algorithm estimates the amount of solutions for a given Grover oracle.

    Parameters
    ----------
    qv : QuantumVariable
        The QuantumVariable on which to evaluate.
    oracle : function
        The oracle function.
    precision : int
        The precision to perform the quantum phase estimation with.

    Returns
    -------
    M : float
        An estimate of the amount of solutions.

    Examples
    --------

    We create an oracle, which performs a simple phase flip on the last qubit. ::

        from qrisp import quantum_counting, z, QuantumVariable

        def oracle(qv):
            z(qv[-1])


    We expect half of the state-space of the input to be a solution.

    For 3 qubits, the state space is $2^3 = 8$ dimensional.

    >>> quantum_counting(QuantumVariable(3), oracle, 3)
    3.999999999999999

    For 4 qubits, the state space is $2^4 = 16$ dimensional.

    >>> quantum_counting(QuantumVariable(4), oracle, 3)
    7.999999999999998


    """

    from qrisp import gate_wrap
    from qrisp.grover import diffuser

    @gate_wrap
    def grover_operator(qv):
        oracle(qv)
        diffuser(qv)

    h(qv)
    res = QPE(qv, grover_operator, precision=precision)

    mes_res = res.get_measurement()

    theta = min(list(mes_res.keys())[:1]) * 2 * np.pi

    N = 2**qv.size
    M = N * np.sin(theta / 2) ** 2

    return M


def HHL(qv, hamiltonian_evolution, ev_inversion, precision):
    """
    Evaluates the HHL algorithm.

    Parameters
    ----------
    qv : TYPE
        DESCRIPTION.
    hamiltonian_evolution : TYPE
        DESCRIPTION.
    ev_inversion : TYPE
        DESCRIPTION.
    precision : TYPE
        DESCRIPTION.

    Returns
    -------
    ancilla : TYPE
        DESCRIPTION.

    """
    from qrisp import QuantumBool, invert

    # Perform quantum phase estimation
    res = QPE(qv, hamiltonian_evolution, precision)

    ancilla = QuantumBool()

    ev_inversion(res, ancilla)

    # Perform quantum phase estimation inverse
    with invert():
        QPE(qv, hamiltonian_evolution, target=res)

    res.delete()

    return ancilla


def fredkin_qc(num_ctrl_qubits=1, ctrl_state=-1, method="gray"):
    from qrisp import QuantumCircuit, XGate
    mcx_gate = XGate().control().control(ctrl_state=ctrl_state, method=method)

    qc = QuantumCircuit(num_ctrl_qubits + 2)
    qc.cx(-1, -2)
    qc.append(mcx_gate, qc.qubits)
    qc.cx(-1, -2)

    return qc


def demux(input, ctrl_qv, output=None, ctrl_method=None, permit_mismatching_size=False, reduce_depth = False):
    """
    This functions allows moving an input value into an iterable output, where the
    position is specified by a ``QuantumFloat``. Demux is short for demultiplexer and
    is a standard component in `classical electrical circuitry
    <https://en.wikipedia.org/wiki/Multiplexer>`_.

    Demux can either move qubit states into a QuantumVariable or ``QuantumVariables``
    into ``QuantumArrays``.

    This function can also be used to "in-place demux" the 0-th entry of an iterable to
    the position specified by ``ctrl_qv``. For more information on this, check the
    second example.

    Parameters
    ----------
    input : Qubit or QuantumVariable
        The input value that is supposed to be moved.
    ctrl_qv : QuantumFloat
        The QuantumFloat specifying to which output the input should be moved.
    output : QuantumVariable or QuantumArray, optional
        The output object, where the input should end up. By default, a new object
        (QuantumVariable or QuantumArray) is created. Note that when this parameter is
        given, it is guaranteed, that the 0-th entry will be moved to the desired
        position, the other entries can also be permuted away from their original
        position.
    ctrl_method : string, optional
        The ``ctrl_method`` string passed to the
        :ref:`control environment <ControlEnvironment>` to generate controlled swaps.
    permit_mismatching_size : bool, optional
        If set to False, an exception will be raised, if the state-space dimension of
        `ctrl_qv`` is differing from the amount of outputs. The default is False.
    reduce_depth : bool, optional
        If set to True, this option reduces (de)allocates additional qubits to
        reduce the depth. The default is False.

    Raises
    ------
    Exception
        Tried to demux with mismatchingly sized control input.

    Returns
    -------
    output : QuantumVariable or QuantumArray
        The output object with the input signal placed at the index specified by
        ``ctrl_qv``.

    Examples
    --------

    We create a ``QuantumBool`` and demux it into a ``QuantumArray`` ::

        from qrisp import *

        qb = QuantumBool()
        qb.flip()

        index = QuantumFloat(2)

        h(index[1])

        res_array = demux(qb, index)

    >>> print(multi_measurement([index, res_array]))
    {(0, OutcomeArray([1., 0., 0., 0.])): 0.5, (2, OutcomeArray([0., 0., 1., 0.])): 0.5}

    Demux can also be used to move the 0-th entry of a ``QuantumArray`` in-place. ::

        qa = QuantumArray(shape = 4, qtype = qb)

        qa[0].flip()

        demux(qa[0], index, qa)

    >>> print(multi_measurement([index, qa]))
    {(0, OutcomeArray([1., 0., 0., 0.])): 0.5, (2, OutcomeArray([0., 0., 1., 0.])): 0.5}

    For low-level manipulations, demux can move information within ``QuantumVariables``.
    ::

        qf = QuantumVariable(4)

        qf[:] = "1000"

        demux(qf[0], index, qf)

    >>> print(multi_measurement([index, qf]))
    {(0, '1000'): 0.5, (2, '0010'): 0.5}
    """

    from qrisp import QuantumArray, QuantumVariable, Qubit, control, swap

    if output is None:
        if isinstance(input, QuantumVariable):
            output = QuantumArray(input, 2 ** len(ctrl_qv))
        elif isinstance(input, Qubit):
            output = QuantumVariable(2 ** len(ctrl_qv))
        else:
            raise Exception("Don't know how to handle input type " + str(type(input)))
    else:
        if isinstance(output, QuantumArray):
            for qv in output.flatten()[1:]:
                if qv.name == input.name:
                    raise Exception(
                        "Tried to in-place demux QuantumArray entry,"
                        "which is not a 0-th position"
                    )
        elif isinstance(output, QuantumVariable):
            for qb in output.reg[1:]:
                if qb.identifier == input.identifier:
                    raise Exception(
                        "Tried to in-place demux QuantumVariable entry,"
                        "which is not a 0-th position"
                    )

    n = int(np.ceil(np.log2(len(output))))
    N = 2**n

    if len(output) != 2 ** len(ctrl_qv) and not permit_mismatching_size:
        raise Exception("Tried to demux with mismatching sized control input")

    if hash(input) != hash(output[0]):
        swap(input, output[0])

    if not len(ctrl_qv):
        return output

    if len(output) > 2 ** (len(ctrl_qv) - 1):
        with control(ctrl_qv[-1], ctrl_method=ctrl_method):
            swap(output[0], output[N // 2])
    else:
        demux(
            output[0],
            ctrl_qv[:-1],
            output,
            ctrl_method=ctrl_method,
            permit_mismatching_size=permit_mismatching_size,
        )
        return output

    if n > 1:
        
        if reduce_depth:
            demux_ancilla = QuantumVariable(len(ctrl_qv)-1)
            cx(ctrl_qv[:-1], demux_ancilla)
            ctrl_qubits = list(demux_ancilla)
        else:
            ctrl_qubits = ctrl_qv[:-1]
            
        
        demux(
            output[0],
            ctrl_qubits,
            #ctrl_qv[:-1],
            output[: N // 2],
            ctrl_method=ctrl_method,
            permit_mismatching_size=permit_mismatching_size,
            reduce_depth=reduce_depth
        )
        
        
        demux(
            output[N // 2],
            ctrl_qv[:-1],
            output[N // 2 :],
            ctrl_method=ctrl_method,
            permit_mismatching_size=permit_mismatching_size,
            reduce_depth=reduce_depth
        )
        if reduce_depth:
            cx(ctrl_qv[:-1], demux_ancilla)
            demux_ancilla.delete()
        


    return output


def q_indexing(q_array, index):
    from qrisp import invert

    with invert():
        demux(q_array[0], index, q_array, ctrl_method="gray_pt")

    res = q_array[0].duplicate(init=True)

    demux(q_array[0], index, q_array, ctrl_method="gray_pt")

    return res


def q_swap_into(q_array, index, qv):
    from qrisp import invert, swap

    with invert():
        demux(q_array[0], index, q_array, ctrl_method="gray_pt")

    swap(q_array[0], qv)

    demux(q_array[0], index, q_array, ctrl_method="gray_pt")



def cyclic_shift(iterable, shift_amount = 1):
    r"""
    Performs a cyclic shift of the values of an iterable with logarithmic depth. 
    The shifting amount can be specified.

    
    Parameters
    ----------
    iterable : list[Qubit] or list[QuantumVariable] or QuantumArray
        The iterable to be shifted.
    shift_amount : integer or QuantumFloat, optional
        The iterable will be shifted by that amount. The default is 1.

    Examples
    --------
    
    We create a QuantumArray, initiate a sequence of increments and perform a cyclic shift.
    
    >>> from qrisp import QuantumFloat, QuantumArray, cyclic_shift
    >>> import numpy as np
    >>> qa = QuantumArray(QuantumFloat(3), 8)
    >>> qa[:] = np.arange(8)
    >>> cyclic_shift(qa, shift_amount = 2)
    >>> print(qa)
    {OutcomeArray([6, 7, 0, 1, 2, 3, 4, 5]): 1.0}
    
    We do something similar to demonstrate the shift by quantum values.
    For this we initiate a :ref:`QuantumFloat` in the superposition of 0, 1 and -3.
    
    >>> shift_amount = QuantumFloat(3, signed = True)
    >>> shift_amount[:] = {0 : 3**-0.5, 1: 3**-0.5, -3 : 3**-0.5}
    >>> qa = QuantumArray(QuantumFloat(3), 8)
    >>> qa[:] = np.arange(8)
    >>> cyclic_shift(qa, shift_amount)
    >>> print(qa)
    {OutcomeArray([0, 1, 2, 3, 4, 5, 6, 7]): 0.3333, OutcomeArray([7, 0, 1, 2, 3, 4, 5, 6]): 0.3333, OutcomeArray([3, 4, 5, 6, 7, 0, 1, 2]): 0.3333}
    """
    
    from qrisp import QuantumFloat, control, QuantumBool, cx
    
    if isinstance(shift_amount, QuantumFloat):
        
        if shift_amount.mshape[0] < 0:
            raise Exception("Tried to quantum shift by non-integer QuantumFloat")
        
        if shift_amount.signed:
            with control(shift_amount.sign()):
                cyclic_shift(iterable, -2**(shift_amount.mshape[1]))
            
        for i in range(*shift_amount.mshape):
            with control(shift_amount.significant(i)):
                cyclic_shift(iterable, 2**i)
    
        return
            
    
    
    if len(iterable) == 0 or not shift_amount%len(iterable):
        return
    if shift_amount < 0:
        return cyclic_shift(iterable[::-1], -shift_amount)
    
    if shift_amount != 1:
        
        perm = np.arange(len(iterable))
        perm = (perm - shift_amount)%(len(iterable))
        
        permute_iterable(iterable, perm)
        return
    
    n = np.log2(len(iterable))
    
    if n != np.floor(n):
        n = int(np.floor(n))
        
        cyclic_shift(iterable[:2**n], shift_amount)
        cyclic_shift([iterable[0]] + list(iterable[2**n:]), shift_amount)
        return
    
    
    correction_indices = []
    for i in range(len(iterable)//2):
        swap_tuple = (2*i, 2*i+1)
        swap(iterable[swap_tuple[0]], iterable[swap_tuple[1]])
        correction_indices.append(swap_tuple[0])
    
    cyclic_shift([iterable[i] for i in correction_indices], shift_amount)
    

def to_cycles(perm):
    pi = {i: perm[i] for i in range(len(perm))}
    cycles = []

    while pi:
        elem0 = next(iter(pi)) # arbitrary starting element
        this_elem = pi[elem0]
        next_item = pi[this_elem]

        cycle = []
        while True:
            cycle.append(this_elem)
            del pi[this_elem]
            this_elem = next_item
            if next_item in pi:
                next_item = pi[next_item]
            else:
                break

        cycles.append(cycle)

    return cycles

def permute_iterable(iterable, perm):
    """
    Applies an arbitrary permutation to an iterable with logarithmic depth.

    Parameters
    ----------
    iterable : QuantumArray or List[QuantumVariable] or QuantumVariable or List[Qubit]
        The iterable to perform the permutation on.
    perm : List[integer]
        A list specifying the permutation.

    Examples
    --------
    
    We create a QuantumArray containing increments and apply a specified permutation.
    
    >>> from qrisp import QuantumFloat, QuantumArray, permute_iterable
    >>> import numpy as np
    >>> qa = QuantumArray(QuantumFloat(3), 8)
    >>> qa[:] = np.arange(8)
    >>> permute_iterable(qa, perm = [1,0,3,7,5,2,6,4])
    >>> print(qa)
    {OutcomeArray([1, 0, 3, 7, 5, 2, 6, 4]): 1.0}
    >>> print(qa.qs)
    QuantumCircuit:
    ---------------
      qa.0: ────────────X──────────────────────
                        │                      
      qa.1: ──────X─────┼──────────────────────
                  │     │                      
      qa.2: ──────┼──X──┼──────────────────────
            ┌───┐ │  │  │                      
    qa_1.0: ┤ X ├─┼──┼──X──────────────────────
            └───┘ │  │                         
    qa_1.1: ──────X──┼─────────────────────────
                     │                         
    qa_1.2: ─────────X─────────────────────────
    <BLANKLINE>                                           
    qa_2.0: ───────────────────────────X───────
            ┌───┐                      │       
    qa_2.1: ┤ X ├──────────────────────┼──X────
            └───┘                      │  │    
    qa_2.2: ───────────────────────────┼──┼──X─
            ┌───┐                      │  │  │ 
    qa_3.0: ┤ X ├──────────X───────────┼──┼──┼─
            ├───┤          │           │  │  │ 
    qa_3.1: ┤ X ├──────────┼──X────────┼──┼──┼─
            └───┘          │  │        │  │  │ 
    qa_3.2: ───────────────┼──┼──X─────┼──┼──┼─
                           │  │  │     │  │  │ 
    qa_4.0: ──────X────────┼──┼──┼─────┼──┼──┼─
                  │        │  │  │     │  │  │ 
    qa_4.1: ──────┼──X─────┼──┼──┼─────┼──┼──┼─
            ┌───┐ │  │     │  │  │     │  │  │ 
    qa_4.2: ┤ X ├─┼──┼──X──┼──┼──┼─────┼──┼──┼─
            ├───┤ │  │  │  │  │  │     │  │  │ 
    qa_5.0: ┤ X ├─X──┼──┼──┼──┼──┼──X──X──┼──┼─
            └───┘    │  │  │  │  │  │     │  │ 
    qa_5.1: ─────────X──┼──┼──┼──┼──┼──X──X──┼─
            ┌───┐       │  │  │  │  │  │     │ 
    qa_5.2: ┤ X ├───────X──┼──┼──┼──┼──┼──X──X─
            └───┘          │  │  │  │  │  │    
    qa_6.0: ───────────────┼──┼──┼──┼──┼──┼────
            ┌───┐          │  │  │  │  │  │    
    qa_6.1: ┤ X ├──────────┼──┼──┼──┼──┼──┼────
            ├───┤          │  │  │  │  │  │    
    qa_6.2: ┤ X ├──────────┼──┼──┼──┼──┼──┼────
            ├───┤          │  │  │  │  │  │    
    qa_7.0: ┤ X ├──────────X──┼──┼──X──┼──┼────
            ├───┤             │  │     │  │    
    qa_7.1: ┤ X ├─────────────X──┼─────X──┼────
            ├───┤                │        │    
    qa_7.2: ┤ X ├────────────────X────────X────
            └───┘                              
    Live QuantumVariables:
    ----------------------
    QuantumFloat qa
    QuantumFloat qa_1
    QuantumFloat qa_2
    QuantumFloat qa_3
    QuantumFloat qa_4
    QuantumFloat qa_5
    QuantumFloat qa_6
    QuantumFloat qa_7
    
    """
    
    from sympy.combinatorics import Permutation
    
    inv_perm = list(Permutation(perm)**-1)
    
    cycles = to_cycles(inv_perm)
    
    for c in cycles:
        cyclic_shift([iterable[i] for i in c], 1)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import gate_wrap, x


@gate_wrap(is_qfree=True, permeability="args")
def q_int_div(numerator, divisor, adder="cuccaro", n=None, log_output=True):
    # This function performs integer division based on the following algorithm
    # Which is a reformulation of the division algorithm presented in
    # https://en.wikipedia.org/wiki/Division_algorithm#Non-restoring_division

    # Q = 2**(n)-1
    # D *= 2**(n-1)
    # R = int(N)

    # import numpy as np
    # from qrisp.misc import int_as_array

    # for i in range(n-1, -1, -1):

    #     D = D/2
    #     if not np.sign(D)*np.sign(R) != -1:
    #         Q -= 2**(i)

    #     if int_as_array(Q, n)[::-1][i]:
    #         R = R-D
    #     else:
    #         R = R+D

    # Q -= 2**(n-1)
    # Q += 0.5

    # if D < 0:
    #     Q += -0.5
    #     R = R - D
    # else:
    #     Q += -0.5
    #     R = R + D

    # D *= 2

    # The D multiplications/divisions are realized by bitshifts
    # The conditioned Q -= 2**i operation is performed by NOTing the appropriate
    # qubit controlled on the condition

    # The R = R+-D operation is realized by wrapping applying NOT gates on the R
    # qubits before and after the addition controlled on the corresponding Q-qubit
    # This yields the desired behavior because (x' + y)' = x - y
    # where the prime denotes negation

    # The Q -= 2**(n-1) is simply performed by flipping the significance n-1 bit
    # The Q = Q + 0.5 is performed by adding an aditional qubit at the -1 significance
    # of Q and turning it on

    # The final conditional is performed similar to the previous condition
    # Where the variables which are being operated on are wrapped in CNOTs

    qs = numerator.qs

    if numerator.exponent < 0 or divisor.exponent < 0:
        raise Exception(
            "Tried to call integer division with non integer quantum floats"
        )

    # n is the bitsize of the quotient
    # The largest possible number the quotient can take is
    # 2**(numerator_max_sig - divisor_min_sig)
    # We need one additional bit (why?)

    if n is None:
        n = int(numerator.mshape[1] - divisor.exponent) + 1

    # The quotient bits will successively be calculated/
    # added starting at the most significant bit.
    # As the algorithm produces empty bits in the remainder at the same rate as
    # quotient bits are required, we initialize the quotient as a 1 bit variable
    # Since the highest significance ist calculated fist, this bit needs significance n.

    # We initialize this float with a single Qubit which will be freed up
    # once the algorithm produced the first bit
    # We can't initialize variables with 0 qubits
    from qrisp.qtypes.quantum_float import QuantumFloat

    quotient = QuantumFloat(1, n, signed=True)

    # In order to determine the datashape of the remainder, we note that
    # within the algorithm, there are steps where the remainder variable needs to hold
    # larger+finer numbers than the final result for the remainder

    # The crucial step is the addition
    # remainder += divisor

    # Regarding the exponent of the remainder, we know that it has to be
    # less than the exponent of the numerator (so we can properly initialize)
    # But also less than the divisor exponent - 1 in order to properly execute
    # the last iteration which is a D/2 addition
    remainder_exponent = min(numerator.exponent, divisor.exponent - 1)

    # For the maximum significance,
    # we note that the divisor is initially bit shifted by n
    divisor.exp_shift(n - 1)

    # Therefore, the required maximum significance is acquired
    # with the formula for addition max_sig_add = max(max_sig_a, max_sig_b) + 1
    remainder_max_sig = max(numerator.mshape[1], divisor.mshape[1]) + 1

    # Create remainder float
    remainder_size = remainder_max_sig - remainder_exponent
    remainder = QuantumFloat(remainder_size, remainder_exponent, signed=True)

    # We now can initialize the remainder as the value of the numerator
    remainder.init_from(numerator)

    # If the divisor is signed we need to flip the quotient sign bit (why?)
    if divisor.signed:
        qs.cx(divisor[-1], quotient[-1])

    from qrisp.arithmetic import inpl_add

    # Perform iterations
    for i in range(n - 1, -1, -1):
        if log_output:
            R = list(remainder.get_measurement().keys())[0]
            print("R: ", R)

            D = list(divisor.get_measurement().keys())[0]
            print("D: ", D)

            Q = list(quotient.get_measurement().keys())[0]
            print("Q: ", Q)

            # print("Q: " + bin_rep(int(Q/2.**quotient.exponent),
            # quotient.size-1) + "(" + str(Q) + ")")
            print("---")

        # The sign bit of the remainder is the newly calculated bit
        # of the quotient of this iteration
        # Therefore we need to move this bit from the remainder to the quotient
        # We do this in a very "hacky" fashion

        # Remove from remainder
        remainder_sign_bit = remainder.reg.pop(-1)
        remainder.size -= 1
        remainder.mshape[1] -= 1

        # Add to quotient at position 0
        quotient.reg.insert(0, remainder_sign_bit)
        quotient.size += 1
        quotient.mshape[1] += 1

        # This next instruction is a bit involved to understand

        # If we didn't use the technique of transfering the bits from the remainder
        # to the quotient, we would have to call quotient.x() in order to realize
        # the initial Q = 2**n - 1 statement. In this case we would then afterwards
        # CNOT from remainder sign bit it the quotient sign bit from this iteration
        # Since x gate on the quotient qubit and the CNOT commute,
        # we can also perform the x gate after the CNOT

        # Translating this to our case we replaced the CNOT with the bit transfer,
        # we arive at the neccessity of an x at this point
        x(quotient[0])

        # Since the new bit was added at the least significant end,
        # we need to lower the exponent
        quotient.exp_shift(-1)

        # This is to handle the D part of the statement "np.sign(D)*np.sign(R) != 1"
        if divisor.signed:
            qs.cx(divisor[-1], quotient[0])

        # We constructed Q with one initial qubit, which will be freed up after
        # the first iteration. This does not produce a qubit overhead,
        # because the upcomming add function needs more than one ancilla anyway
        if i == n - 1:
            quotient.reduce(quotient[1])

        # Perform the D = D/2 instruction
        divisor.exp_shift(-1)

        # We wrap the remainder in CNOT gates to make use of
        # (R' + D)' = (R - D)
        # where the prime denotes the negation
        for j in range(remainder.size):
            qs.cx(quotient[0], remainder[j])

        # This perform the addition R += D or R -= D
        inpl_add(
            remainder,
            divisor,
            ignore_rounding_error=True,
            ignore_overflow_error=True,
            adder=adder,
        )

        # Instead of executing this layer of CNOT gates,
        # we use the layer from the next iteration and only
        # CNOT the control qubit from this layer
        # for j in range(remainder.size):
        # qs.cx(quotient[0], remainder[j])

        # We still need to make up for the fact that the next layer of CNOT
        # gates doesnt include remainder[-1]

        # qs.cx(quotient[0], remainder[-1])

        # However remainder[-1] is also the control qubit of the next iteration
        # implying our plan of CNOTting the next control qubit cancels the
        # previous line

        # qs.cx(quotient[0], remainder[-1])

        # In any iteration apart from the first, quotient[0] does not contain
        # the desired value because of the sheenanigans of the previous lines.
        # In detail: It contains the desired value (+) the quotient[0] qubit
        # of the previous iteration. We cancel with this command
        if i != n - 1:
            qs.cx(quotient[1], quotient[0])

    # Since there is no more upcoming iteration, we simply perform the CNOT layer
    # without any further tricks.
    for j in range(remainder.size):
        qs.cx(quotient[0], remainder[j])

    # This performs the Q -= 2**(n-1) instruction (the -1 qubit is the sign bit =>
    # the -2 qubit is the significance n-1 bit)
    x(quotient[-2])

    # Handle the case that the divisor is signed
    if divisor.signed:
        # Create Qubit to account for the fact that we will add 0.5 to the quotient,
        # even though it only supports integers
        quotient.extend(1, position=0)
        quotient.exp_shift(-1)
        x(quotient[0])

        # Add/Subtract 0.5 executed via (Q' + 0.5)' = Q - 0.5
        qs.cx(divisor[-1], quotient.reg)
        quotient.incr(-(2**quotient.exponent))
        qs.cx(divisor[-1], quotient.reg)

        # Remove 0.5 significance qubit (contains 0 now)
        quotient.reduce(quotient[0])
        quotient.exp_shift(1)

        # Incase the divisor is not only signed but actually negative, we CNOT the
        # remainder such that the upcoming addition circuit results in a subtraction
        qs.cx(divisor[-1], remainder.reg)

    # Perform R = R +- D
    inpl_add(
        remainder,
        divisor,
        ignore_rounding_error=True,
        ignore_overflow_error=True,
        adder=adder,
    )

    if divisor.signed:
        # Undo the CNOT
        pass
        qs.cx(divisor[-1], remainder.reg)

    # Perform D = 2*D
    divisor.exp_shift(1)

    # Finally, if the numerator was signed, we flip the quotient sign bit
    if numerator.signed:
        qs.cx(numerator[-1], quotient[-1])

    return quotient, remainder


# Wrapper for the integer division
# Supports division for quantum float of arbitrary exponent.
# And allows to give a precision threshold (prec)
# The resulting quotient variable Q_res has the property |Q_res - Q_real| < 2**prec
def q_divmod(numerator, divisor, adder="thapliyal", prec=0):
    """
    Performs division up to arbitrary precision. Returns the quotient and the remainder.

    Parameters
    ----------
    numerator : QuantumFloat
        The QuantumFloat to divide.
    divisor : QuantumFloat
        The QuantumFloat to divide by.
    adder : str, optional
        The type of adder to use. Available are "thapliyal" and "cuccarro".
        The default is "thapliyal".
    prec : int, optional
        The precision of the division. If the precision is set to $k$,
        the approximated quotient $q_{apr}$ and the true quotient $q_{true}$
        satisfy $|q_{apr} - q_{true}|<2^{-k}$. The default is 0.

    Returns
    -------
    quotient : QuantumFloat
        The approximated quotient.
    remainder : QuantumFloat
        The remainder which satisfying $q*d + r = n$.

    Examples
    --------

    We calculate 10/8 with varying precision:

    >>> from qrisp import QuantumFloat, q_divmod, multi_measurement
    >>> num = QuantumFloat(4)
    >>> div = QuantumFloat(4)
    >>> num[:] = 10
    >>> div[:] = 8
    >>> quotient, remainder = q_divmod(num, div, prec = 1)
    >>> multi_measurement([quotient, remainder])
    {(1.0, 2.0): 1.0}

    Now with higher precision

    >>> num = QuantumFloat(4)
    >>> div = QuantumFloat(4)
    >>> num[:] = 10
    >>> div[:] = 8
    >>> quotient, remainder = q_divmod(num, div, prec = 3)
    >>> multi_measurement([quotient, remainder])
    {(1.25, 0.0): 1.0}

    """

    # The idea is to bit shift numerator and divisor by s such that,
    # they are both integers. To increase the precision we shift N by -prec
    # (we'll see shortly why this increases the precision).
    # We write

    # N_tilde = 2**(s-prec) * N
    # D_tilde = 2**s * D

    # We then perform integer division giving Q_tilde, R_tilde such that

    # N_tilde/D_tilde = Q_tilde + R_tilde/D_tilde
    # where |R_tilde| < |D_tilde|

    # We now set

    # Q = Q_tilde 2**(prec)
    # and insert, which yields

    # N_tilde/D_tilde = 2**(-prec) * N/D  = 2**(-prec) * Q + R_tilde/D_tilde

    # =>  N/D = Q + 2**prec*R_tilde/D_tilde

    # So the error is
    # |2**prec*R_tilde/D_tilde| < 2**prec

    # Because R_tilde < D_tilde

    prec = -prec
    s = -min(numerator.exponent, divisor.exponent)

    num_exp_shift = s - prec
    div_exp_shift = s

    numerator.exp_shift(num_exp_shift)
    divisor.exp_shift(div_exp_shift)

    # print("N_tilde: ", numerator.get_measurement())
    # print("D_tilde: ", divisor.get_measurement())

    quotient, remainder = q_int_div(numerator, divisor, adder=adder, log_output=False)

    # print("Q_tilde: ", quotient.get_measurement())
    # print("R_tilde: ", remainder.get_measurement())

    quotient.exp_shift(prec)
    remainder.exp_shift(-num_exp_shift)

    divisor.exp_shift(-div_exp_shift)
    numerator.exp_shift(-num_exp_shift)

    return quotient, remainder


def q_div(numerator, divisor, prec=None):
    """
    Performs division up to arbitrary precision and uncomputes the remainder.

    Parameters
    ----------
    numerator : QuantumFloat
        The QuantumFloat to divide.
    divisor : QuantumFloat
        The QuantumFloat to divide by.
    prec : int, optional
        The precision of the division. If the precision is set to $k$,
        the approximated quotient $q_{apr}$ and the true quotient $q_{true}$
        satisfy $|q_{apr} - q_{true}|<2^{-k}$.
        By default, a suited precision will be determined from the other inputs.

    Returns
    -------
    quotient : QuantumFloat
        The result of the division.

    Examples
    --------

    We calculate 10/8:

    >>> from qrisp import QuantumFloat, q_div
    >>> num = QuantumFloat(4)
    >>> div = QuantumFloat(4)
    >>> num[:] = 10
    >>> div[:] = 8
    >>> quotient = q_div(num, div, prec = 2)
    >>> print(quotient)
    {1.25: 1.0}

    """

    from qrisp import U_g_inpl_adder, h, hybrid_mult

    if prec is None:
        prec = divisor.size - numerator.exponent

    quotient, remainder = q_divmod(numerator, divisor, prec=prec)

    hybrid_mult(quotient, divisor, remainder, init_op="qft", terminal_op=None)

    U_g_inpl_adder(remainder, numerator, mult_factor=-1)

    h(remainder)
    # QFT(remainder, inv = True, exec_swap = False)
    # remainder += -1

    remainder.delete()

    return quotient


def qf_inversion(qf, prec=None):
    """
    Calculates the multiplicative inverse of a QuantumFloat.

    Parameters
    ----------
    qf : QuantumFloat
        The QuantumFloat to invert.
    prec : int, optional
        The precision of the inversion. If the precision is set to $k$,
        the approximated inverse $q_{apr}$ and the true inverse $q_{true}$ satisfy
        $|q_{res} - q_{true}|<2^{-k}$.
        By default, a suited precision will be determined from the other input.

    Returns
    -------
    result : QuantumFloat
        A QuantumFloat containing the inverse.

    Examples
    --------

    We calculate the inverse of 0.75

    >>> from qrisp import QuantumFloat, qf_inversion
    >>> qf = QuantumFloat(2, -2)
    >>> qf[:] = 0.75
    >>> qf_inv = qf_inversion(qf, prec = 4)
    >>> print(qf_inv)
    {1.3125: 1.0}
    >>> 0.75**-1
    1.3333333333333333

    """

    if prec is None:
        prec = qf.size

    from qrisp import QuantumFloat, x

    numerator = QuantumFloat(1, signed=False)
    numerator.encode(1)

    result = q_div(numerator, qf, prec)

    x(numerator)
    numerator.delete()

    return result



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


def test_QuantumArray_indexing():
    from qrisp import QuantumBool, QuantumArray, QuantumFloat, h, multi_measurement

    q_array = QuantumArray(QuantumBool(), shape=(4, 4))
    index_0 = QuantumFloat(2)
    index_1 = QuantumFloat(2)

    index_0[:] = 2
    index_1[:] = 1

    h(index_0[0])

    with q_array[index_0, index_1] as entry:
        entry.flip()

    assert q_array.most_likely()[index_0.most_likely(), index_1.most_likely()] == True


def test_QuantumDictionary_load():
    from qrisp import QuantumDictionary, QuantumFloat, h

    qtype = QuantumFloat(4, -2)
    float_qd = QuantumDictionary(return_type=qtype)
    float_qd[0] = 1
    float_qd[1] = 2

    key_qv = QuantumFloat(1, signed=False)
    h(key_qv)
    value_qv = qtype.duplicate()

    float_qd.load(key_qv, value_qv, synth_method="pprm")
    assert value_qv.get_measurement() == {1.0: 0.5, 2.0: 0.5}


def test_QuantumFloat():
    from qrisp import QuantumFloat, invert

    qf_6 = QuantumFloat(3)
    qf_7 = QuantumFloat(3)
    qf_6[:] = 7
    qf_7[:] = 2
    qf_8 = qf_6 / qf_7
    assert qf_8.get_measurement() == {3.5: 1.0}

    qf_9 = qf_6 // qf_7
    assert qf_9.get_measurement() == {3.0: 1.0}

    qf_10 = QuantumFloat(3, -1)
    qf_10[:] = 3.5
    qf_11 = qf_10**-1
    assert qf_11.get_measurement() == {0.25: 1.0}

    qf_15 = QuantumFloat(3)
    qf_15[:] = 4
    qf_16 = QuantumFloat(3)
    qf_16[:] = 3
    qf_15 += qf_16
    qf_15 -= 2
    assert qf_15.get_measurement() == {5.0: 1.0}


def test_ConditionEnvironment():
    from qrisp import (
        QuantumChar,
        QuantumFloat,
        QuantumBool,
        QuantumVariable,
        h,
        x,
        cx,
        mcx,
        p,
        multi_measurement,
        ConditionEnvironment,
    )
    import numpy as np

    q_ch = QuantumChar()
    qf = QuantumFloat(3, signed=True)

    h(q_ch[0])

    with q_ch == "a":
        qf += 2

    with q_ch == "a" as cond_bool:
        qf += 2
        cond_bool.flip()
        qf -= 2
        p(np.pi, cond_bool)

    assert multi_measurement([q_ch, qf]) == {("a", 4): 0.5, ("b", -2): 0.5}

    def quantum_eq(qv_0, qv_1):
        res = QuantumBool(name="res_bool")

        if qv_0.size != qv_1.size:
            raise Exception(
                "Tried to evaluate equality condition for QuantumVariables of differing size"
            )

        temp_qv = QuantumVariable(qv_0.size)

        cx(qv_0, temp_qv)
        cx(qv_1, temp_qv)
        x(temp_qv)

        mcx(temp_qv, res)

        return res

    # Create some sample arguments on which to evaluate the condition
    q_bool_0 = QuantumBool()
    q_bool_1 = QuantumBool()
    q_bool_2 = QuantumBool()

    h(q_bool_0)

    with ConditionEnvironment(cond_eval_function=quantum_eq, args=[q_bool_0, q_bool_1]):
        q_bool_2.flip()

    assert multi_measurement([q_bool_0, q_bool_1, q_bool_2]) == {
        (False, False, True): 0.5,
        (True, False, False): 0.5,
    }


def test_InversionEnvironment():
    from qrisp import QuantumFloat, invert, QuantumBool, h, q_mult, multi_measurement

    qf = QuantumFloat(3)

    h(qf)

    mult_res = q_mult(qf, qf)

    q_bool = QuantumBool()

    with mult_res < 10:
        q_bool.flip()

    with invert():
        q_mult(qf, qf, target=mult_res)

    mult_res.delete(verify=True)

    assert multi_measurement([qf, q_bool]) == {
        (0, True): 0.125,
        (1, True): 0.125,
        (2, True): 0.125,
        (3, True): 0.125,
        (4, False): 0.125,
        (5, False): 0.125,
        (6, False): 0.125,
        (7, False): 0.125,
    }



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.simulator.bi_arrays import (
    BiArray,
    DenseBiArray,
    DummyBiArray,
    SparseBiArray,
    tensordot,
)
from qrisp.simulator.circuit_reordering import *
from qrisp.simulator.impure_quantum_state import ImpureQuantumState
from qrisp.simulator.quantum_state import QuantumState, TensorFactor
from qrisp.simulator.simulator import *
from qrisp.simulator.unitary_management import *



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import weakref

# This module contains the necessary tools to merge QuantumSessions

# Due to the interplay with QuantumEnvironments and the behavior
# that QuantumEnvironments automatically detect QuantumSessions that operated
# within them there is quite a bit of complexity involved here.

# The idea to merge two session qs_0 and qs_1 is to first add the appropriate
# amount of qubits/clbits to qs_0, append all the instructions and
# fix the other attributes Then we let all the attributes of qs_1 point
# to the corresponding attributes of qs_0 such that any change on one quantum session
# also appears in the other quantum session

# Therefore, two QuantumSessions are declared as identical (or merged) if their
# attributes (like .data, .qubits) point to the same lists as the other one.
# The == operator for QuantumSessions checks if the .data attribute points to the same
# list.


# If a merge is requested, the first step is to "engage" the QuantumSessions in the
# environment stack. By default, QuantumEnvironments create their own QuantumSession
# and "engaging" here means that if merge is called on a QuantumSession within an
# environment, this QuantumSession will be merged with all the QuantumSessions of the
# environment stack.
def merge_sessions(qs_0, qs_1):
    # We now iterate through every QuantumEnvironment and insert its data into the
    # correct original_data attribute
    for qs in [qs_0, qs_1]:
        # We need to find the environment where the env_qs quantum session is not merged
        # into qs. This implies that the instructions of qs have been appended in this
        # environment's parent. Therefore, all the data needs to go into the
        # original_data attribute of this environment.
        for env in qs.env_stack:
            if not env.env_qs == qs:
                env.original_data.extend(qs.data)
                qs.data = []
                merge_sessions_inner(qs, env.env_qs)

    merge_sessions_inner(qs_0, qs_1)


def merge_sessions_inner(qs_0, qs_1, merge_env_stack_=True):
    if qs_0 == qs_1:
        return

    from qrisp.quantum_network import QuantumNetworkSession

    if isinstance(qs_0, QuantumNetworkSession):
        if qs_0.backend is not qs_1.backend:
            if isinstance(qs_1, QuantumNetworkSession):
                raise Exception(
                    "Tried to merge two QuantumNetworkSessions containing "
                    "differing clients"
                )

            if qs_1.backend is not None:
                raise Exception(
                    "Tried to merge a QuantumNetworkSession with a "
                    "QuantumSession with a non-trivial backend"
                )

        qs_1.backend = qs_0.backend
        qs_1.inbox = qs_0.inbox
        qs_1.__class__ = qs_0.__class__

    if isinstance(qs_1, QuantumNetworkSession):
        if qs_1.backend is not qs_0.backend:
            if isinstance(qs_0, QuantumNetworkSession):
                raise Exception(
                    "Tried to merge two QuantumNetworkSessions containing "
                    "differing clients"
                )

            if qs_0.backend is not None:
                raise Exception(
                    "Tried to merge a QuantumNetworkSession with a "
                    "QuantumSession with a non-trivial backend"
                )

        qs_0.backend = qs_1.backend
        qs_0.inbox = qs_1.inbox
        qs_0.__class__ = qs_1.__class__

    # The problem we face here is the following:
    # If two sessions a, b are merged and b is merged with another session c at a later
    # point, the attributes of b are updated by that merge, but the attributes of a are
    # not. To solve this problem, we declare one of the merge participants the "parent"
    # of the merge. The parent has an attribute called "shadow_sessions" that contain
    # the sessions that the parent has been merged with. The attribute updates are then
    # performed on all shadow sessions.

    # To make sure we only operate on parent sessions, we check the parent attribute
    while hasattr(qs_0, "parent"):
        qs_0 = qs_0.parent

    while hasattr(qs_1, "parent"):
        qs_1 = qs_1.parent

    resolve_naming_collisions(qs_0, qs_1)

    intersecting_qubits = set([qb.identifier for qb in qs_0.qubits]).intersection(
        [qb.identifier for qb in qs_1.qubits]
    )

    if len(intersecting_qubits):
        raise Exception(
            f"Tried to merge two QuantumSessions containing identically named "
            f"qubits {intersecting_qubits}"
        )

    if len(qs_0.env_stack) < len(qs_1.env_stack):
        qs_0, qs_1 = qs_1, qs_0

    if merge_env_stack_:
        merge_env_stack(qs_0, qs_1)

        if qs_0 == qs_1:
            return

    # The qs_tracker list should only contain one Session instance per equivalence
    # class, i.e. if two sessions are merged only one of them should appear in the
    # qs_tracker
    i = 0
    while i < len(qs_0.qs_tracker):
        if qs_0.qs_tracker[i]() is None:
            qs_0.qs_tracker.pop(i)
        elif qs_0.qs_tracker[i]() == qs_1:
            qs_0.qs_tracker.pop(i)
        else:
            i += 1

    # Each environment contains a list of QuantumSessions that when it was entered.
    # This list determines if a QuantumSession is engaged into an QuantumEnvironment, if
    # merge is called from inside the QuantumEnvironment. We only want to have one of
    # the two QuantumSessions that are merged in this list, which is resolved by the
    # identify_sessions_in_environment
    for i in range(len(qs_1.env_stack)):
        identify_sessions_in_environment(qs_1.env_stack[-i - 1], qs_0, qs_1)

    if qs_0.backend is not None and qs_1.backend is not None:
        if id(qs_0.backend) != id(qs_1.backend):
            raise Exception(
                "Tried to merge QuantumSessions with differing, "
                "non-trivial default backends."
            )

    if qs_0.backend is None:
        qs_0.backend = qs_1.backend

    if set(qs_0.clbits).intersection(qs_1.clbits):
        raise Exception("Tried to merge sessions with common classical bits")

    for qb in qs_1.qubits:
        qb.qs = weakref.ref(qs_0)
    for cb in qs_1.clbits:
        cb.qs = weakref.ref(qs_0)

    qs_0.qubits.extend(qs_1.qubits)
    qs_0.clbits.extend(qs_1.clbits)

    qs_0.data.extend(qs_1.data)

    object.__setattr__(qs_1, "data", qs_0.data)

    # Update qv_list
    while len(qs_1.qv_list):
        qv = qs_1.qv_list.pop()
        # Patch quantum session attribute
        qv.qs = qs_0

        qs_0.qv_list.append(qv)

    for qv in qs_1.deleted_qv_list:
        # Patch quantum session attribute
        qv.qs = qs_0
        
    qs_0.deleted_qv_list.extend(qs_1.deleted_qv_list)

    reorder_quantum_variables(qs_0)

    # Add variables to the uncomputation stack
    qs_0.uncomp_stack.extend(qs_1.uncomp_stack)

    qs_0.shadow_sessions.extend(qs_1.shadow_sessions + [weakref.ref(qs_1)])

    i = 0
    while i < len(qs_0.shadow_sessions):
        qs = qs_0.shadow_sessions[i]()
        if qs is None:
            qs_0.shadow_sessions.pop(i)
            continue

        qs.parent = qs_0

        # Update all the attributes
        for var in qs_0.__dict__.keys():
            qs.__dict__[var] = qs_0.__dict__[var]

        i += 1


def resolve_naming_collisions(qs_0, qs_1):
    qs_names_0 = [qv.name for qv in qs_0.qv_list + qs_0.deleted_qv_list]
    qs_names_1 = [qv.name for qv in qs_1.qv_list + qs_1.deleted_qv_list]

    intersecting_qv_names = set(qs_names_0).intersection(qs_names_1)

    if len(intersecting_qv_names):
        for qv_name in intersecting_qv_names:
            qv_0_index = qs_names_0.index(qv_name)

            qv_0 = (qs_0.qv_list + qs_0.deleted_qv_list)[qv_0_index]

            qv_1_index = qs_names_1.index(qv_name)
            qv_1 = (qs_1.qv_list + qs_1.deleted_qv_list)[qv_1_index]

            if qv_1.user_given_name:
                if qv_0.user_given_name:
                    raise Exception(
                        "Tried to merge QuantumSession containing "
                        f"identically named QuantumVariables {qv_1.name}"
                    )

                qv_0, qv_1 = qv_1, qv_0

            else:
                if qv_0.creation_time > qv_1.creation_time:
                    if not qv_1.user_given_name:
                        qv_0, qv_1 = qv_1, qv_0

            proposed_new_name = qv_1.name + "_1"
            k = 1

            while proposed_new_name in qs_names_0 + qs_names_1:
                k += 1
                proposed_new_name = qv_1.name + "_" + str(k)

            qv_1.name = proposed_new_name

            for k in range(len(qv_1.reg)):
                qv_1.reg[k].identifier = qv_1.name + "." + str(k)


def merge_env_stack(qs_0, qs_1):
    # We need to find the environment where the env_qs quantum session is not merged
    # into qs. This implies that the instructions of this session have been appended in
    # this environment's parent. Therefore, all the data needs to go into the
    # original_data attribute of this environment.
    for env in qs_0.env_stack:
        if env.env_qs == qs_0 and not env.env_qs == qs_1:
            if env in qs_1.env_stack:
                env.original_data.extend(qs_1.data)
                qs_1.data = []

        if env.env_qs == qs_1 and not env.env_qs == qs_0:
            for env_ in qs_0.env_stack:
                if not env_.env_qs == qs_0:
                    env_.original_data.extend(qs_0.data)
                    qs_0.data = []
                    merge_sessions_inner(qs_0, env_.env_qs, False)
                if env.env_qs == env_.env_qs:
                    break


def reorder_quantum_variables(qs):
    overall_qv_list = qs.qv_list + qs.deleted_qv_list

    qs.qv_list.sort(key=lambda x: x.creation_time)

    overall_qv_list.sort(key=lambda x: x.creation_time)

    sorted_qubit_list = []
    for qv in overall_qv_list:
        sorted_qubit_list += qv.reg

    sorted_qubit_list.extend(list(set(qs.qubits) - set(sorted_qubit_list)))

    qs.qubits = sorted_qubit_list


def identify_sessions_in_environment(env, qs_0, qs_1):
    i = 0
    while i < len(env.active_qs_list):
        if env.active_qs_list[i]() is None:
            env.active_qs_list.pop(i)
            continue
        elif env.active_qs_list[i]() == qs_1:
            env.active_qs_list.pop(i)
            continue
        i += 1


def multi_session_merge(input_list):
    session_list = input_list
    session_list.sort(key=lambda x: -len(x.data))

    if not len(session_list):
        from qrisp.core import QuantumSession

        return QuantumSession()

    for i in range(1, len(session_list)):
        merge_sessions(session_list[0], session_list[i])

    return session_list[0]


def recursive_qs_search(input):
    if isinstance(input, str):
        return []

    from qrisp.core import QuantumSession

    if hasattr(input, "__iter__"):
        iterable = True

    # This case would in principle allow for also searching objects
    # this however requires a considerable chunk of resources
    elif hasattr(input, "__dict__") and not isinstance(input, QuantumSession):
        iterable = False
        # input = input.__dict__
    else:
        iterable = False

    result = []
    if iterable:
        if isinstance(input, dict):
            for key in input.keys():
                result += recursive_qs_search(key)
                result += recursive_qs_search(input[key])
        else:
            for i in range(len(input)):
                result += recursive_qs_search(input[i])
    else:
        if isinstance(input, QuantumSession):
            result = [input]
        else:
            try:
                if isinstance(input(), QuantumSession):
                    result = [input()]
            except TypeError:
                pass

            try:
                if isinstance(input.qs(), QuantumSession):
                    result = [input.qs()]
            except AttributeError:
                pass

    return result


def recursive_qv_search(input):
    if isinstance(input, str):
        return []
    from qrisp.core import QuantumVariable

    if hasattr(input, "__iter__"):
        iterable = True
    elif hasattr(input, "__dict__") and not isinstance(input, QuantumVariable):
        iterable = False
        # input = input.__dict__
    else:
        iterable = False

    result = []
    if iterable:
        if isinstance(input, dict):
            for key in input.keys():
                result += recursive_qv_search(key)
                result += recursive_qv_search(input[key])
        else:
            for i in range(len(input)):
                result += recursive_qv_search(input[i])
    else:
        if isinstance(input, QuantumVariable):
            result = [input]

    return result


def merge(*args):
    session_list = recursive_qs_search(args)
    return multi_session_merge(session_list)
